'''
This part contains actions filterings in the filtering widget.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSlot
from lib import sqlitedb


class MainFrameFilterListSlots:

    #######################################################################
    #                          Filter list slots                          #
    #######################################################################

    @pyqtSlot(QtWidgets.QListWidgetItem)
    def filterItemClicked(self, item):
        """Do a doc filtering using a selected value of a selected type

        Args:
            item (QListWidgetItem): selected item in the filter list widget.

        This is a slot to the filter_item_list.itemClicked signal.
        """

        self.logger.info('Clicked filter item.text() = %s' %item.text())

        filter_type=self.filter_type_combbox.currentText()
        filter_text=item.text()
        current_folder=self._current_folder
        if current_folder:
            folderid=current_folder[1]

            filter_docids=sqlitedb.filterDocs(self.meta_dict,self.folder_data,
                    filter_type,filter_text,folderid)

            if len(filter_docids)>0:
                self.loadDocTable(None,filter_docids,sortidx=4,sel_row=0)

            sel=self.filter_type_combbox.currentText()

            if sel=='Filter by keywords':
                self.clear_filter_label.setText(
                        'Showing documents with keyword "%s"' %filter_text)
            elif sel=='Filter by authors':
                self.clear_filter_label.setText(
                        'Showing documents authored by "%s"' %filter_text)
            elif sel=='Filter by publications':
                self.clear_filter_label.setText(
                        'Showing documents published in "%s"' %filter_text)
            elif sel=='Filter by tags':
                self.clear_filter_label.setText(
                        'Showing documents tagged "%s"' %filter_text)

            self.clear_filter_frame.setVisible(True)

        return


    @pyqtSlot()
    def filterTypeCombboxChange(self):
        """Change filter type and populate filter values

        This is a slot to the filter_type_combbox.currentIndexChanged signal.
        And is called everytime a folder is selected. See clickSelFolder().
        keywords, authors, publications or tags in a folder (not desending
        into sub-folders) are collected, depending on the selected filter
        type, and entries are added to the filter list.

        """

        # clear current filtering first
        self.clearFilterButtonClicked()

        # remove duplicate frame
        self.clearDuplicateButtonClicked()

        # remove search result frame
        self.clearSearchResButtonClicked()

        sel=self.filter_type_combbox.currentText()
        current_folder=self._current_folder

        self.logger.info('Filter type combobox select = %s' %sel)

        if current_folder:

            self.logger.info('current_folder = %s, folderid_d = %s'\
                    %(current_folder[0], current_folder[1]))

            #---------------Get items in folder---------------
            foldername,folderid=current_folder
            if folderid=='-1':
                docids=list(self.meta_dict.keys())
            else:
                docids=self.folder_data[folderid]

            if sel=='Filter by keywords':
                folderdata=sqlitedb.fetchMetaData(self.meta_dict,'keywords_l',docids,
                        unique=True,sort=True)
            elif sel=='Filter by authors':
                folderdata=sqlitedb.fetchMetaData(self.meta_dict,'authors_l',docids,
                        unique=False,sort=False)
                # why not unique and sort?
            elif sel=='Filter by publications':
                folderdata=sqlitedb.fetchMetaData(self.meta_dict,'publication',docids,
                        unique=True,sort=True)
            elif sel=='Filter by tags':
                folderdata=sqlitedb.fetchMetaData(self.meta_dict,'tags_l',docids,
                        unique=True,sort=True)

            folderdata=list(set(folderdata))
            folderdata.sort()
            self.filter_item_list.clear()
            self.filter_item_list.addItems(folderdata)

        return


    @pyqtSlot()
    def clearFilterButtonClicked(self):
        '''Hide the filter result header frame above the doc table

        This is a slot to the clicked signal of the clear_filter_button
        shown in the clear_filter_frame.

        It is also called in filterTypeCombboxChange(), which is called
        on selecting a folder. Therefore selecting/switching to a folder
        will automatically hide the clear_filter_frame.
        '''

        if not self.doc_table.isVisible():
            self.doc_table.setVisible(True)

        if self.clear_filter_frame.isVisible():
            self.clear_filter_frame.setVisible(False)

            current_folder=self._current_folder
            if current_folder:
                folder,folderid=current_folder

                # TODO: keep a record of previous sortidx?
                if folder=='All' and folderid=='-1':
                    self.loadDocTable(None,sortidx=4,sel_row=0)
                else:
                    self.loadDocTable((folder,folderid),sortidx=4,sel_row=0)

        return



'''
Defines getters for the current states of various widgets.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from lib import sqlitedb


class MainFrameProperties:

    @property
    def _tabledata(self):
        if hasattr(self,'doc_table'):
            return self.doc_table.model().arraydata
        else:
            return None

    @property
    def _current_doc(self):
        if hasattr(self,'doc_table'):
            if len(self._tabledata)==0:
                return None
            current_row=self.doc_table.currentIndex().row()
            if current_row < len(self._tabledata):
                docid=self._tabledata[current_row][0]
                return docid
            else:
                self.logger.critical('_current_doc: current row > row number. current_row = %s, len(self._tabledata) = %d' %(current_row, len(self._tabledata)))
                return None
        else:
            return None

    @property
    def _current_meta_dict(self):
        if hasattr(self,'t_meta'):
            if hasattr(self.t_meta,'fields_dict'):
                return self.t_meta.fields_dict
        return None

    @property
    def _current_folder_item(self):
        if hasattr(self,'libtree'):
            item=self.libtree.selectedItems()
            if item:
                return item[0]
            else:
                return None
        return None

    @property
    def _current_folder(self):
        if hasattr(self,'libtree'):
            item=self._current_folder_item
            if item:
                return item.data(0,0), item.data(1,0) # folder name, folderid
        return None

    @property
    def _current_docids(self):
        if hasattr(self,'doc_table'):
            docid=[ii[0] for ii in self._tabledata]
            return docid
        else:
            return None

    @property
    def _trashed_folder_ids(self):
        if hasattr(self,'folder_dict'):
            #return self.libtree._trashed_folder_ids
            return sqlitedb.getTrashedFolders(self.folder_dict)

    @property
    def _orphan_doc_ids(self):
        if hasattr(self,'meta_dict'):
            return [kk for kk in self.meta_dict if self.meta_dict[kk]['deletionPending']=='true']



'''
Preference dialog.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import logging
from collections import OrderedDict
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, QPoint, pyqtSlot, QRect
from PyQt5.QtGui import QFont, QPainter
from PyQt5.QtWidgets import QStyle, QStyleOptionSlider, QDialogButtonBox
from ..tools import getHLine, getXMinYExpandSizePolicy

LOGGER=logging.getLogger(__name__)



class PreferenceDialog(QtWidgets.QDialog):

    def __init__(self,settings,parent=None):
        '''
        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py
        '''

        super(PreferenceDialog,self).__init__(parent=parent)

        self.settings=settings
        self.parent=parent

        self.font_dict=OrderedDict([
            ('Meta Tab -> Title'      , 'display/fonts/meta_title')    ,
            ('Meta Tab -> Authors'    , 'display/fonts/meta_authors')  ,
            ('Meta Tab -> Keywords'   , 'display/fonts/meta_keywords') ,
            ('Document Table Entries' , 'display/fonts/doc_table')     ,
            ('Bibtex Tab'             , 'display/fonts/bibtex')        ,
            ('Notes Tab'              , 'display/fonts/notes')         ,
            ('Scratch Pad Tab'        , 'display/fonts/scratch_pad')
            ])

        self.new_values={} # store changes before applying

        self.label_color='color: rgb(0,0,140); background-color: rgb(235,235,240)'
        self.title_label_font=QFont('Serif',12,QFont.Bold)
        self.sub_title_label_font=QFont('Serif',10,QFont.Bold)

        self.resize(800,600)
        self.setWindowTitle('Preferences')
        self.setWindowModality(Qt.ApplicationModal)

        v_layout=QtWidgets.QVBoxLayout()
        h_layout=QtWidgets.QHBoxLayout()
        #h_layout.setContentsMargins(10,40,10,20)
        self.setLayout(v_layout)

        title_label=QtWidgets.QLabel('    Change Preference Settings')
        title_label.setFont(QFont('Serif',12,QFont.Bold))
        v_layout.addWidget(title_label)

        v_layout.addLayout(h_layout)

        self.cate_list=QtWidgets.QListWidget(self)
        #self.list.setSizePolicy(getXMinYExpandSizePolicy())
        self.cate_list.setMaximumWidth(150)
        h_layout.addWidget(self.cate_list)

        #self.cate_list.setStyleSheet('''
            #QListWidget::item { border: 0px solid rgb(235,235,240);
            #font: 14px;
            #background-color: rgb(205,205,245);
            #color: rgb(100,10,13) };
            #background-color: rgb(230,234,235);
            #''')

        self.cate_list.addItems(['Citation Style', 'Display', 'Bibtex Export',
            'RIS Export', 'Savings', 'Miscellaneous'])

        self.content_vlayout=QtWidgets.QVBoxLayout()
        h_layout.addLayout(self.content_vlayout)

        self.buttons=QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Apply |\
                    QDialogButtonBox.Cancel,
            Qt.Horizontal, self)

        self.buttons.accepted.connect(lambda: (self.applyChanges(), self.accept()))
        self.buttons.rejected.connect(self.reject)
        self.buttons.button(QDialogButtonBox.Apply).clicked.connect(
                self.applyChanges)

        self.content_vlayout.addWidget(self.buttons)

        self.cate_list.currentItemChanged.connect(self.cateSelected)
        self.cate_list.setCurrentRow(0)


    @pyqtSlot()
    def applyChanges(self):
        '''Write changes to settings'''

        LOGGER.info('Changes: %s' %self.new_values)

        for kk,vv in self.new_values.items():
            self.settings.setValue(kk,vv)

        #------------------Set new timer------------------
        if 'saving/auto_save_min' in self.new_values:
            interval=self.settings.value('saving/auto_save_min',1,int)
            self.parent.main_frame.auto_save_timer.setInterval(interval*60*1000)
            LOGGER.info('Set auto save timer to %s' %interval)

        self.new_values={}

        #---------------Create output folder---------------
        storage_folder=self.settings.value('saving/storage_folder')
        if not os.path.exists(storage_folder):
            os.makedirs(storage_folder)

            LOGGER.info('Create new storage folder %s' %storage_folder)

        # TODO: apply change to database and meta_dict
        # need to call saveFoldersToDatabase() with new folder, and
        # metaDictToDatabase() for all docids. Then move database file over.
        # may need to require a reboot if so.

        #sqlitedb.saveFoldersToDatabase(self.db,self.folder_dict,
                #self.settings.value('saving/storage_folder'))
        # UPDATE: after changing the file paths relative, seems no need to
        # do the above things.

        return


    @pyqtSlot(QtWidgets.QListWidgetItem)
    def cateSelected(self,item):
        '''Load widgets for a selected category

        Args:
            item (QListWidgetItem): selected category item.
        '''

        item_text=item.text()
        LOGGER.info('item.text() = %s' %item.text())

        if self.content_vlayout.count()>1:
            self.content_vlayout.removeWidget(self.content_frame)

        if item_text=='Display':
            self.content_frame=self.loadDisplayOptions()
        elif item_text=='Bibtex Export':
            self.content_frame=self.loadBibExportOptions()
        elif item_text=='RIS Export':
            self.content_frame=self.loadRISExportOptions()
        elif item_text=='Citation Style':
            self.content_frame=self.loadCitationStyleOptions()
        elif item_text=='Savings':
            self.content_frame=self.loadSavingsOptions()
        elif item_text=='Miscellaneous':
            self.content_frame=self.loadMiscellaneousOptions()

        self.content_vlayout.insertWidget(0,self.content_frame)


    def createFrame(self,title):
        '''Create a template frame for a category page

        Args:
            title (str): title of the category

        Returns:
            scroll (QScrollArea): a scroll area.
            va (QVBoxLayout): the vertical box layout used in scroll.
        '''

        frame=QtWidgets.QWidget(self)
        scroll=QtWidgets.QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setWidget(frame)
        va=QtWidgets.QVBoxLayout()
        frame.setLayout(va)
        va.setSpacing(int(va.spacing()*2))

        label=QtWidgets.QLabel(title)
        label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)
        va.addWidget(label)
        #va.addWidget(getHLine(self))

        return scroll, va


    def createOmitKeyGroup(self):
        '''Create checkbox group
        '''

        grid=QtWidgets.QGridLayout()

        self.groupbox=QtWidgets.QGroupBox('Omit these fields in the exported bib entry.')
        self.groupbox.setCheckable(True)

        omittable_keys=[
            'read', 'favourite', 'added', 'confirmed', 'firstNames_l',
            'lastName_l', 'deletionPending', 'folders_l', 'type', 'id',
            'abstract', 'advisor', 'month', 'language', 'confirmed',
            'deletionPending', 'note', 'publicLawNumber', 'sections',
            'reviewedArticle', 'userType', 'shortTitle', 'sourceType',
            'code', 'codeNumber', 'codeSection', 'codeVolume', 'citationKey',
            'day', 'dateAccessed', 'internationalAuthor', 'internationalUserType',
            'internationalTitle', 'internationalNumber', 'genre', 'lastUpdate',
            'legalStatus', 'length', 'medium'
            ]
        omittable_keys.sort()

        omit_keys=self.settings.value('export/bib/omit_fields', [], str)
        # likely something wrong with qt. When list is set empty by
        # settings.setValue('key',[]), on the NEXT load of the program,
        # settings.value('export/bib/omit_fields', [], str) gives ''
        if isinstance(omit_keys,str) and omit_keys=='':
            omit_keys=[]

        for ii,keyii in enumerate(omittable_keys):
            checkboxii=QtWidgets.QCheckBox(keyii,self.groupbox)
            if keyii in omit_keys:
                checkboxii.setChecked(True)
            checkboxii.stateChanged.connect(self.omitKeyChanged)
            grid.addWidget(checkboxii,int(ii/3),ii%3)

        self.groupbox.toggled.connect(lambda on: self.omitKeysGroupChanged(on,
            self.groupbox))

        self.groupbox.setLayout(grid)

        return self.groupbox


    def loadDisplayOptions(self):
        '''Load widgets for the Display category'''

        scroll,va=self.createFrame('Select Fonts')

        ha=QtWidgets.QHBoxLayout()
        #ha.addStretch()
        label=QtWidgets.QLabel('NOTE some changes requires re-booting.')
        ha.addWidget(label,0,Qt.AlignTop)

        text_list=QtWidgets.QListWidget()
        text_list.setSizePolicy(getXMinYExpandSizePolicy())
        text_list.addItems(self.font_dict.keys())

        ha.addWidget(text_list)

        text_list.itemDoubleClicked.connect(self.chooseFont)
        va.addLayout(ha)

        return scroll


    @pyqtSlot(QtWidgets.QListWidgetItem)
    def chooseFont(self,item):
        '''Choose a font

        Args:
            item (QListWidgetItem): item for field font change is applying to
        '''

        item_text=item.text()
        LOGGER.info('item.text() = %s' %item.text())

        font_setting_name=self.font_dict[item_text]
        default=self.settings.value(font_setting_name, QFont)
        new_font,isok=QtWidgets.QFontDialog.getFont(default,
                caption='Choose Font for %s' %item_text)

        if isok:
            self.new_values[font_setting_name]=new_font
            LOGGER.info('Font after change = %s' %new_font)

        return


    def loadSavingsOptions(self):
        '''Load widgets for the Savings category'''

        scroll, va=self.createFrame('Rename Files')

        #-------------Choose storage folder section-------------
        # NOTE: too much trouble dealing the folder path changes. Not now
        """
        label2=QtWidgets.QLabel('''
        Select folder to save document files. <br/>
        &nbsp;&nbsp; Document (e.g. PDFs) will be copied to the
        <span style="font:bold;">
        "Collections" </span> <br/> &nbsp;&nbsp; sub-folder of the chosen folder.
        ''')
        label2.setTextFormat(Qt.RichText)
        va.addWidget(label2)

        ha=QtWidgets.QHBoxLayout()
        ha.addStretch()

        storage_folder=self.settings.value('saving/storage_folder')
        le=QtWidgets.QLineEdit()
        le.setText(storage_folder)

        va.addWidget(le)
        va.addLayout(ha)
        button=QtWidgets.QPushButton(self)
        button.setText('Choose')

        button.clicked.connect(self.chooseSaveFolder)
        ha.addWidget(button)
        va.addWidget(getHLine())
        """

        #---------------Rename file section---------------
        checkbox=QtWidgets.QCheckBox('Rename Files')
        checked=self.settings.value('saving/rename_files',type=int)
        LOGGER.debug('Is rename files = %s' %checked)

        checkbox.setChecked(checked)

        le=QtWidgets.QLineEdit(self)
        le.setReadOnly(True)
        le.setDisabled(1-checked)
        checkbox.stateChanged.connect(lambda on: self.changeRenameFiles(on))
        checkbox.stateChanged.connect(lambda on: le.setEnabled(on))

        le.setText('Renaming Format: Author_Year_Title.pdf')

        label2=QtWidgets.QLabel('''
        Documents (e.g. PDFs) will be copied to the
        <span style="font:bold;">
        "%s/_collections" </span>
        folder, and renamed by the following format.
        ''' %(self.settings.value('saving/current_lib_folder')))
        label2.setTextFormat(Qt.RichText)
        label2.setWordWrap(True)

        va.addWidget(checkbox)
        va.addWidget(label2)
        va.addWidget(le)

        #----------------Auto save section----------------
        va.addWidget(getHLine(self))
        label3=QtWidgets.QLabel('Auto save interval (min)')
        label3.setStyleSheet(self.label_color)
        label3.setFont(self.title_label_font)
        va.addWidget(label3)
        va.addWidget(getHLine(self))

        slider=LabeledSlider(1,10,1,parent=self)
        slider.sl.setValue(self.settings.value('saving/auto_save_min',2,int))
        slider.sl.valueChanged.connect(self.changeSavingInterval)
        slider.setMaximumWidth(400)

        va.addWidget(slider)

        va.addStretch()


        return scroll


    def chooseSaveFolder(self):
        '''Prompt for dir choose and store new value'''

        fname=QtWidgets.QFileDialog.getExistingDirectory(self,
            'Choose a folder to save documents and database')

        if fname:
            LOGGER.info('Folder after change = %s' %fname)
            self.new_values['saving/storage_folder']=fname

        return


    def changeRenameFiles(self,on):
        '''Store the check state of rename file checkbox'''

        on=1 if on>0 else 0 # for some reason <on> keeps giving me 2
        self.new_values['saving/rename_files']=on
        LOGGER.info('Change rename files to %s' %on)

        return


    def changeSavingInterval(self,value):
        '''Store the value on the auto saving interval slider'''

        LOGGER.info('Change auto saving interval to %s' %value)
        self.new_values['saving/auto_save_min']=value

        return


    def loadBibExportOptions(self):
        '''Load widgets for the bibtex Export category'''

        scroll, va=self.createFrame('Bibtex Export')
        self.groupbox=self.createOmitKeyGroup()
        va.addWidget(self.groupbox)

        va.addWidget(getHLine())

        label=QtWidgets.QLabel('File path type')
        label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)

        self.radio_groupbox=self.createPathTypeGroup('bib')
        va.addWidget(self.radio_groupbox)

        return scroll


    def loadRISExportOptions(self):
        '''Load widgets for the RIS Export category'''

        scroll, va=self.createFrame('RIS Export')

        label=QtWidgets.QLabel('File path type')
        label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)

        self.radio_groupbox=self.createPathTypeGroup('ris')
        va.addWidget(self.radio_groupbox)

        va.addStretch()

        return scroll


    def createPathTypeGroup(self, export_type):
        '''Create radiobutton group for file path type selection'''

        groupbox=QtWidgets.QGroupBox('Use relative or absolute paths for files.')
        ha=QtWidgets.QHBoxLayout()
        path_type=self.settings.value('export/%s/path_type' %export_type,str)

        for ii in ['Relative', 'Absolute']:
            buttonii=QtWidgets.QRadioButton(ii)
            if ii.lower()==path_type:
                buttonii.setChecked(True)
            buttonii.toggled.connect(lambda on: self.pathTypeChanged(on, export_type))
            ha.addWidget(buttonii)

        groupbox.setLayout(ha)

        return groupbox


    @pyqtSlot(bool, str)
    def pathTypeChanged(self,on, export_type):
        '''Collect check states in the path type radiobutton group'''

        LOGGER.debug('export_type = %s' %export_type)
        for rii in self.radio_groupbox.findChildren(QtWidgets.QRadioButton):
            if rii.isChecked():
                self.new_values['export/%s/path_type' %export_type]=rii.text().lower()
                break

        return


    def omitKeyChanged(self,on):
        '''Store changes in the omit keys checkbox group'''

        self.new_values['export/bib/omit_fields']=self.getOmitKeys()

        return


    def omitKeysGroupChanged(self, on, groupbox):
        '''Change check states in the omit keys checkbox group as a whole'''

        omit_keys=[]

        for box in groupbox.findChildren(QtWidgets.QCheckBox):
            box.stateChanged.disconnect()
            box.setChecked(on)
            box.setEnabled(True)
            box.stateChanged.connect(self.omitKeyChanged)
            if box.isChecked():
                omit_keys.append(box.text())

        self.new_values['export/bib/omit_fields']=omit_keys

        return


    def getOmitKeys(self):
        '''Collect check states in the omit key checkbox group'''

        omit_keys=[]

        for box in self.groupbox.findChildren(QtWidgets.QCheckBox):
            if box.isChecked():
                omit_keys.append(box.text())

        return omit_keys


    def loadCitationStyleOptions(self):
        '''Load widgets for the Citaiton Style category'''

        scroll, va=self.createFrame('Citation Styles')
        va.addStretch()

        return scroll


    def loadMiscellaneousOptions(self):
        '''Load widgets for the Miscellaneous category'''

        scroll, va=self.createFrame('Auto Open')

        #-------Open last database on launch section-------
        checkbox=QtWidgets.QCheckBox('Automatically Open Last Database on Start-up?')
        checkbox.stateChanged.connect(self.changeAutoOpenLast)
        auto_open_last=self.settings.value('file/auto_open_last',type=int)
        if auto_open_last==1:
            checkbox.setChecked(True)
        else:
            checkbox.setChecked(False)

        va.addWidget(checkbox)
        va.addWidget(getHLine(self))

        #--------------Recent number section--------------
        label1=QtWidgets.QLabel('Number of Recently Opened Database')
        label1.setStyleSheet(self.label_color)
        label1.setFont(self.title_label_font)
        va.addWidget(label1)

        slider2=LabeledSlider(0,10,1,parent=self)
        slider2.sl.setValue(self.settings.value('file/recent_open_num',type=int))
        slider2.sl.valueChanged.connect(self.changeRecentNumber)
        slider2.setMaximumWidth(400)

        va.addWidget(slider2)
        va.addWidget(getHLine())

        #------------Duplicate check min score------------
        label2=QtWidgets.QLabel('Duplicate Check')
        label2.setStyleSheet(self.label_color)
        label2.setFont(self.title_label_font)
        va.addWidget(label2)

        label3=QtWidgets.QLabel('Minimum Similarity Score to Define Duplicate (1-100)')
        self.spinbox=QtWidgets.QSpinBox()
        self.spinbox.setMinimum(1)
        self.spinbox.setMaximum(100)
        self.spinbox.setValue(self.settings.value('duplicate_min_score',type=int))
        self.spinbox.valueChanged.connect(self.changeDuplicateMinScore)

        ha=QtWidgets.QHBoxLayout()
        ha.addWidget(label3)
        ha.addWidget(self.spinbox)

        va.addLayout(ha)
        va.addStretch()

        return scroll


    def changeAutoOpenLast(self,on):
        '''Store the check state of the auto open checkbox'''

        if on:
            self.new_values['file/auto_open_last']=1
        else:
            self.new_values['file/auto_open_last']=0

        LOGGER.info('Change auto open last to %s' %on)

        return


    def changeRecentNumber(self,value):
        '''Store the value on the recent open slider'''

        LOGGER.info('Change recent database number to %s' %value)
        self.new_values['file/recent_open_num']=value

        return


    def changeDuplicateMinScore(self,value):
        '''Store the value in the duplicate minimum score spinbox'''

        LOGGER.info('Change min duplicate score to %s' %value)
        self.new_values['duplicate_min_score']=value

        return


class LabeledSlider(QtWidgets.QWidget):
    def __init__(self, minimum, maximum, interval=1, orientation=Qt.Horizontal,
            labels=None, parent=None):
        '''Slider widget with labels

        Args:
            minimum (int): min value.
            maximum (int): max value.
            interval (int): interval.
        Kwargs:
            orientation (int): orientation.
            labels (str or None): labels.
            parent (QWidget or None): parent widget.
        '''

        super(LabeledSlider, self).__init__(parent=parent)

        levels=range(minimum, maximum+interval, interval)
        if labels is not None:
            if not isinstance(labels, (tuple, list)):
                raise Exception("<labels> is a list or tuple.")
            if len(labels) != len(levels):
                raise Exception("Size of <labels> doesn't match levels.")
            self.levels=list(zip(levels,labels))
        else:
            self.levels=list(zip(levels,map(str,levels)))

        if orientation==Qt.Horizontal:
            self.layout=QtWidgets.QVBoxLayout(self)
        elif orientation==Qt.Vertical:
            self.layout=QtWidgets.QHBoxLayout(self)
        else:
            raise Exception("<orientation> wrong.")

        # gives some space to print labels
        self.left_margin=10
        self.top_margin=10
        self.right_margin=10
        self.bottom_margin=10

        self.layout.setContentsMargins(self.left_margin,self.top_margin,
                self.right_margin,self.bottom_margin)

        self.sl=QtWidgets.QSlider(orientation, self)
        self.sl.setMinimum(minimum)
        self.sl.setMaximum(maximum)
        self.sl.setValue(minimum)
        if orientation==Qt.Horizontal:
            self.sl.setTickPosition(QtWidgets.QSlider.TicksBelow)
            #self.sl.setMinimumWidth(300) # just to make it easier to read
        else:
            self.sl.setTickPosition(QtWidgets.QSlider.TicksLeft)
            #self.sl.setMinimumHeight(300) # just to make it easier to read
        self.sl.setTickInterval(interval)
        self.sl.setSingleStep(1)

        self.layout.addWidget(self.sl)

    def paintEvent(self, e):

        super(LabeledSlider,self).paintEvent(e)

        style=self.sl.style()
        painter=QPainter(self)
        st_slider=QStyleOptionSlider()
        st_slider.initFrom(self.sl)
        st_slider.orientation=self.sl.orientation()

        length=style.pixelMetric(QStyle.PM_SliderLength, st_slider, self.sl)
        available=style.pixelMetric(QStyle.PM_SliderSpaceAvailable, st_slider, self.sl)

        for v, v_str in self.levels:

            # get the size of the label
            rect=painter.drawText(QRect(), Qt.TextDontPrint, v_str)

            if self.sl.orientation()==Qt.Horizontal:
                # I assume the offset is half the length of slider, therefore
                # + length//2
                x_loc=QStyle.sliderPositionFromValue(self.sl.minimum(),
                        self.sl.maximum(), v, available)+length//2

                # left bound of the text = center - half of text width + L_margin
                left=x_loc-rect.width()//2+self.left_margin
                bottom=self.rect().bottom()

                # enlarge margins if clipping
                if v==self.sl.minimum():
                    if left<=0:
                        self.left_margin=rect.width()//2-x_loc
                    if self.bottom_margin<=rect.height():
                        self.bottom_margin=rect.height()

                    self.layout.setContentsMargins(self.left_margin,
                            self.top_margin, self.right_margin,
                            self.bottom_margin)

                if v==self.sl.maximum() and rect.width()//2>=self.right_margin:
                    self.right_margin=rect.width()//2
                    self.layout.setContentsMargins(self.left_margin,
                            self.top_margin, self.right_margin,
                            self.bottom_margin)

            else:
                y_loc=QStyle.sliderPositionFromValue(self.sl.minimum(),
                        self.sl.maximum(), v, available, upsideDown=True)

                bottom=y_loc+length//2+rect.height()//2+self.top_margin-3
                # there is a 3 px offset that I can't attribute to any metric

                left=self.left_margin-rect.width()
                if left<=0:
                    self.left_margin=rect.width()+2
                    self.layout.setContentsMargins(self.left_margin,
                            self.top_margin, self.right_margin,
                            self.bottom_margin)

            pos=QPoint(left, bottom)
            painter.drawText(pos, v_str)

        return




'''
General purpose functions.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import re
import time
import platform
import logging
from functools import reduce
from fuzzywuzzy import fuzz
from PyQt5 import QtWidgets
from PyQt5.QtCore import QThread, QObject, QMutex, pyqtSignal, pyqtSlot, Qt
try:
    from . import sqlitedb
except:
    import sqlitedb

LOGGER=logging.getLogger(__name__)


def getMinSizePolicy():
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum,
            QtWidgets.QSizePolicy.Minimum)
    return sizePolicy


def getXMinYExpandSizePolicy():
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum,
            QtWidgets.QSizePolicy.Expanding)
    return sizePolicy


def getXExpandYMinSizePolicy():
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum)
    return sizePolicy


def getXExpandYExpandSizePolicy():
    sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Expanding)
    return sizePolicy


def getHSpacer():
    h_spacer = QtWidgets.QSpacerItem(0,0,QtWidgets.QSizePolicy.Expanding,
            QtWidgets.QSizePolicy.Minimum)
    return h_spacer


def getVSpacer():
    v_spacer = QtWidgets.QSpacerItem(0,0,QtWidgets.QSizePolicy.Minimum,
            QtWidgets.QSizePolicy.Expanding)
    return v_spacer


def getHLine(parent=None):
    h_line = QtWidgets.QFrame(parent)
    h_line.setFrameShape(QtWidgets.QFrame.HLine)
    h_line.setFrameShadow(QtWidgets.QFrame.Sunken)
    return h_line


def getVLine(parent=None):
    v_line = QtWidgets.QFrame(parent)
    v_line.setFrameShape(QtWidgets.QFrame.VLine)
    v_line.setFrameShadow(QtWidgets.QFrame.Sunken)
    return v_line


class WorkerThread(QThread):
    '''NOT IN USE'''

    jobdone_signal=pyqtSignal()
    def __init__(self,func,jobqueue,outqueue,parent=None):
        QThread.__init__(self,parent)

        self.func=func
        self.jobqueue=jobqueue
        self.outqueue=outqueue

    def __del__(self):
        self.wait()

    def run(self):
        while True:
            print('\n# <WorkThread>: Thread start processing ... Remaining queue size: %d.'\
                %(self.jobqueue.qsize()))
            args=self.jobqueue.get()
            rec=self.func(*args)
            self.jobqueue.task_done()
            self.outqueue.put(rec)
            self.jobdone_signal.emit()
            time.sleep(0.1)


class Worker(QObject):
    '''NOT IN USE'''
    sgnFinished = pyqtSignal()

    def __init__(self, parent, func, jobq, outq):
        QObject.__init__(self, parent)
        self._mutex = QMutex()
        self._running = True
        self.jobq=jobq
        self.outq=outq
        self.func=func

    @pyqtSlot()
    def stop(self):
        print('switching while loop condition to false')
        self._mutex.lock()
        self._running = False
        self._mutex.unlock()

    def running(self):
        try:
            self._mutex.lock()
            return self._running
        finally:
            self._mutex.unlock()

    @pyqtSlot()
    def work(self):
        while self.running():
            args=self.jobq.get()
            rec=self.func(*args)
            self.jobq.task_done()
            self.outq.put(rec)
            #time.sleep(0.1)
            print('doing work...')

        self.sgnFinished.emit()


class Client(QObject):
    '''NOT IN USE'''
    clientDone = pyqtSignal()

    def __init__(self, parent, func, jobq, outq, callback):
        QObject.__init__(self, parent)
        self.jobq=jobq
        self.outq=outq
        self.func=func
        self.callback=callback
        self.clientDone.connect(callback)
        self._thread = QThread()
        self._worker = Worker(None,self.func,self.jobq,self.outq)
        self._worker.sgnFinished.connect(self.on_worker_done)

    def start(self):
        self._thread.started.connect(self._worker.work)
        self._worker.moveToThread(self._thread)
        self._thread.start()

    def stop(self):
        print('stopping the worker object')
        self._worker.stop()

    @pyqtSlot()
    def on_worker_done(self):
        print('workers job was interrupted manually')
        self._thread.quit()
        self.clientDone.emit()


def parseAuthors(authorlist):
    """Parse a list of author names

    Args:
        authorlist (list): list of author names: ['firname, lastname', ...]

    Returns:
        firstnames (list): list of first names.
        lastnames (list): list of last names.
        authorlist (list): list of author names.

    """
    firstnames=[]
    lastnames=[]
    for nii in authorlist:
        nii=nii.split(',',1)
        lastnames.append(nii[0] if len(nii)>1 else nii[0])
        firstnames.append(nii[1] if len(nii)>1 else '')
    #authors=sqlitedb.zipAuthors(firstnames,lastnames)

    return firstnames,lastnames,authorlist


def removeInvalidPathChar(path):
    '''Make dir and remove invalid windows path characters

    ':' is illegal in Mac and windows. Strategy: remove, although legal in Linux.
    '''
    path=os.path.abspath(path)

    if platform.system().lower()=='windows':
        drive,remain=os.path.splitdrive(path)
        remain=re.sub(r'[<>:"|?*]','_',remain)
        remain=remain.strip()
        path=os.path.join(drive,remain)
    else:
        path=re.sub(r'[<>:"|?*]','_',path)

    return path


def fuzzyMatch(jobid, id1, id2, dict1, dict2):
    """Compute similarity score between 2 docs using fuzzy matching

    Args:
        jobid (int): job id.
        id1 (int): id of doc 1.
        id2 (int): id of doc 2.
        dict1 (DocMeta): meta data dict of doc 1.
        dict2 (DocMeta): meta data dict of doc 2.

    Returns:
        rec (int): 0 for success.
        jobid (int): input jobid.
        match_result (tuple): in the format ((id1, id2), score). Where score
                              is an int in [0,100], higher means more similar.

    """

    authors1=dict1['authors_l'] or ''
    authors2=dict2['authors_l'] or ''
    authors1=', '.join(authors1)
    authors2=', '.join(authors2)

    title1=dict1['title'] or ''
    title2=dict2['title'] or ''

    journal1=dict1['publication'] or ''
    journal2=dict2['publication'] or ''
    year1=dict1['year'] or ''
    year2=dict2['year'] or ''

    jy1='%s %s' %(journal1, year1)
    jy2='%s %s' %(journal2, year2)

    ratio_authors=fuzz.token_sort_ratio(authors1, authors2)
    ratio_title=fuzz.ratio(title1, title2)
    ratio_other=fuzz.ratio(jy1, jy2)

    len_authors=0.5*(len(authors1)+len(authors2))
    len_title=0.5*(len(title1)+len(title2))
    len_other=0.5*(len(jy1)+len(jy2))

    score=(len_authors*ratio_authors + len_title*ratio_title + len_other*ratio_other)/\
            (len_authors+len_title+len_other)

    LOGGER.debug('authors1 = %s, authors2 = %s, score = %d'\
            %(authors1, authors2, ratio_authors))
    LOGGER.debug('title1 = %s, title2 = %s, score = %d'\
            %(title1, title2, ratio_title))
    LOGGER.debug('jy1 = %s, jy2 = %s, score = %d'\
            %(jy1, jy2, ratio_other))

    return 0,jobid, ((id1,id2), round(score))


def dfsCC(edges):
    '''Get connected components in undirected graph using DFS

    Args:
        edges (list): list of vertices.

    Returns:
        ccs.values (list): list of connected components, in the format:
            [(v1, v2, ...), (v3, v4, ...) ...],
            each tuple is a connected component.
    '''

    def explore(v,cc):
        visited.append(v)
        if cc not in ccs:
            ccs[cc]=[v,]
        else:
            ccs[cc].append(v)

        for wii in adj_list[v]:
            if wii not in visited:
                explore(wii,cc)

    vertices=list(set(reduce(tuple.__add__,edges)))
    rev_edges=[(v2,v1) for v1,v2 in edges]

    # get adjacency list
    adj_list={}
    for vii in vertices:
        adj_list[vii]=[eii[1] for eii in edges+rev_edges if eii[0]==vii]

    visited=[]
    ccs={}
    cc=0

    for v in vertices:
        if v not in visited:
            explore(v,cc)
            cc+=1

    return list(ccs.values())


def createFolderTree(folder_dict,parent):
    """Create a folder tree using QTreeWidget

    Args:
        folder_dict (dict): folder structure info. keys: folder id in str,
            values: (foldername, parentid) tuple.
        parent (QWidget): parent of QTreeWidget.

    Returns:
        folder_tree (QTreeWidget): QTreeWidget with the folder structure.
    """

    def addFolder(parent,folderid,folder_dict):

        foldername,parentid=folder_dict[folderid]
        fitem=QtWidgets.QTreeWidgetItem([foldername,str(folderid)])
        style=QtWidgets.QApplication.style()
        diropen_icon=style.standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
        fitem.setIcon(0,diropen_icon)
        sub_ids=sqlitedb.getChildFolders(folder_dict,folderid)
        if parentid=='-1':
            fitem.setFlags(fitem.flags() | Qt.ItemIsTristate |\
                    Qt.ItemIsUserCheckable)
            fitem.setCheckState(0, Qt.Unchecked)
            #parent.addTopLevelItem(fitem)
            parent.addChild(fitem)
        else:
            fitem.setFlags(fitem.flags() | Qt.ItemIsUserCheckable)
            fitem.setCheckState(0, Qt.Unchecked)
            parent.addChild(fitem)
        if len(sub_ids)>0:
            for sii in sub_ids:
                addFolder(fitem,sii,folder_dict)

        return

    folder_tree=QtWidgets.QTreeWidget(parent)
    folder_tree.setColumnCount(2)
    folder_tree.setHeaderHidden(True)
    folder_tree.setColumnHidden(1,True)
    folder_tree.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
    folder_tree.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
    folder_tree.header().setStretchLastSection(False)
    folder_tree.header().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeToContents)
    folder_tree.setDragDropMode(QtWidgets.QAbstractItemView.NoDragDrop)

    #---------------------Add All---------------------
    allitem=QtWidgets.QTreeWidgetItem(['All', '-1'])
    style=QtWidgets.QApplication.style()
    diropen_icon=style.standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
    allitem.setIcon(0,diropen_icon)
    allitem.setFlags(allitem.flags() | Qt.ItemIsTristate |\
            Qt.ItemIsUserCheckable)
    allitem.setCheckState(0, Qt.Unchecked)
    folder_tree.addTopLevelItem(allitem)

    #-------------Get all level 1 folders-------------
    folders1=[(vv[0],kk) for kk,vv in folder_dict.items() if\
            vv[1] in ['-1',]]
    folders1.sort()

    #------------Add folders to tree------------
    for fnameii,idii in folders1:
        addFolder(allitem,idii,folder_dict)

    allitem.setExpanded(True)

    return folder_tree


def iterTreeWidgetItems(treewidget, root=None):
    """Iterate through all items in a QTreeWidget

    Args:
        treewidget (QTreeWidget): QTreeWidget to Iterate.
        root (None or QTreeWidgetItem): start point.
    Returns:
        yield each item found.
    """

    if root is None:
        root=treewidget.invisibleRootItem()

    stack = [root]
    while stack:
        parent = stack.pop(0)
        for row in range(parent.childCount()):
            child = parent.child(row)
            yield child
            if child.childCount()>0:
                stack.append(child)


def autoRename(abpath):
    '''Auto rename a file to avoid overwriting an existing file

    Args:
        abpath (str): absolute path to a folder or a file to rename.

    Returns:
        newname (str): new file path.

    If no conflict found, return <abpath>;
    If conflict with existing file, return renamed file path,
    by appending "_(n)".
    E.g.
        n1='~/codes/tools/send2ever.py'
        n2='~/codes/tools/send2ever_(4).py'
    will be renamed to
        n1='~/codes/tools/send2ever_(1).py'
        n2='~/codes/tools/send2ever_(5).py'
    '''

    def rename_sub(match):
        base=match.group(1)
        delim=match.group(2)
        num=int(match.group(3))
        return '%s%s(%d)' %(base,delim,num+1)

    if not os.path.exists(abpath):
        return abpath

    folder,filename=os.path.split(abpath)
    basename,ext=os.path.splitext(filename)
    # match filename
    rename_re=re.compile('''
            ^(.+?)       # File name
            ([- _])      # delimiter between file name and number
            \\((\\d+)\\) # number in ()
            (.*)         # ext
            $''',\
            re.X)

    newname='%s_(1)%s' %(basename,ext)
    while True:
        newpath=os.path.join(folder,newname)

        if not os.path.exists(newpath):
            break
        else:
            if rename_re.match(newname):
                newname=rename_re.sub(rename_sub,newname)
                newname='%s%s' %(newname,ext)
            else:
                raise Exception("Exception")

    newname=os.path.join(folder,newname)
    return newname





'''
Utitlity functions for extract meta data from PDF file.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import re
from pprint import pprint
import logging

try:
    from . import sqlitedb
except:
    import sqlitedb

from collections import Counter

from pdfminer.pdfparser import PDFParser
from pdfminer.pdfdocument import PDFDocument
from PyPDF2 import PdfFileReader


#---------Regex pattern for matching dois---------
DOI_PATTERN=re.compile(r'(?:doi:)?\s?(10.[1-9][0-9]{3}/.*$)',
        re.DOTALL|re.UNICODE)

LOGGER=logging.getLogger(__name__)



def getPDFMeta_pypdf2(path):
    """Extract meta data from PDF file using PyPDF2

    Args:
        path (str): abspath to PDF file.

    Returns: info (dict): meta data dict.
    """

    with open(path, 'rb') as fin:
        pdf = PdfFileReader(fin)
        info = pdf.getDocumentInfo()
        number_of_pages = pdf.getNumPages()

    LOGGER.info('Read PDF file %s. NO. of pages = %d' %(path, number_of_pages))
    LOGGER.debug('Info of PDF file: %s' %info)

    return info


def getPDFMeta_pdfminer(path):
    """Extract meta data from PDF file using pdfminer

    Args:
        path (str): abspath to PDF file.

    Returns: info (dict): meta data dict.
    """

    with open(path, 'rb') as fin:
        parser = PDFParser(fin)
        doc = PDFDocument(parser)

    LOGGER.info('Read PDF file %s.' %path)
    LOGGER.debug('Info of PDF file: %s' %doc.info)

    return doc.info


def parseToList(text):
    """Convert a field sequence in string to a list.

    Args:
        text (str): string sequence.

    Returns: result (list): sequence split into a list.
    """

    result=[]
    if ' and ' in text:
        delim=' and '
    else:
        delim=',' if ',' in text else ';'
    textlist=text.replace('\n',delim).strip(delim).split(delim)
    for tii in textlist:
        tii=str(tii).strip()
        if len(tii)>0:
            result.append(tii)
    return result


def prepareMeta(meta_dict):
    """Format meta data got from PDF file

    Args:
        meta_dict (dict): dict of meta data extracted from PDF.

    Returns: DocMeta dict.

    Format changes done in this func:
        * 'author' is changed to 'authors_l', and value is a list.
        * 'keywords' is changed to 'keywords_l', and value is a list.
        * doi is searched in a regex.
    """

    result=sqlitedb.DocMeta()
    gotdois=[]
    for kk,vv in meta_dict.items():
        if len(vv)>0:
            kk=kk.lower().lstrip('/') # pypdf2 convention starts with /
            if kk=='author':
                result['authors_l']=parseToList(vv)
            elif kk=='keywords':
                result['keywords_l']=parseToList(vv)
            elif kk in ['title', ]:
                result[kk]=str(vv)

            match=DOI_PATTERN.match(vv)
            if match:
                gotdois.append(vv)

    if len(gotdois)>0:
        counter=Counter(gotdois)
        doi=counter.most_common(1)[0][0]
        result['doi']=doi

    return result



if __name__=='__main__':

    FILE_IN='testpdf2.pdf'
    FILE_IN='mypdf2.pdf'

    print('\nget meta data using pdfminer--------------')
    m1=getPDFMeta_pdfminer(FILE_IN)

    print('\nget meta data using pypdf2--------------')
    m2=getPDFMeta_pypdf2(FILE_IN)

    #result=sqlitedb.DocMeta()
    #result.update(m1[0])

    aa=prepareMeta(m2)
    print('\nmeta from pdf:')
    pprint(aa)



'''
Utitlity functions for annotation extraction.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from datetime import datetime
from PyPDF2.generic import *



YELLOW = [0.95, 0.9, 0.2]



def getRGBcolor(color,verbose=True):
    '''Convert html color to RGB (0,1)
    '''
    color=color.replace('#','')
    rgb=tuple(map(ord,color.decode('hex')))
    return [float(ii)/255 for ii in rgb]


def floatArray(lst):
    return ArrayObject([FloatObject(i) for i in lst])


def now():
    # Python timezone handling is a messs, so just use UTC
    return TextStringObject(datetime.utcnow().\
            strftime("D:%Y%m%d%H%M%SZ00'00"))


def _baseAnno(rect, contents=None, author=None, subject=None,
                       cdate=None, color=None, alpha=1, flag=4):
    '''Set shared properties of all markup annotations.'''

    if cdate is None:
        cdate = now()
    else:
        assert isinstance(cdate, datetime),\
                "cdate is not a datetime object"
        cdate = TextStringObject(cdate.strftime("D:%Y%m%d%H%M%SZ00'00"))

    retval = DictionaryObject({\
            NameObject('/CA'): FloatObject(alpha),\
            NameObject('/F'): NumberObject(flag),\
            NameObject('/Rect'): floatArray(rect),\
            NameObject('/Type'): NameObject('/Annot'),\
            NameObject('/CreationDate'): cdate,\
            NameObject('/M'): now(),\
            })

    # Whether to add an explicit popup when adding to page
    retval.popup = False
    if contents is not None:
        retval[NameObject('/Contents')] = TextStringObject(contents)
    if author is not None:
        retval[NameObject('/T')] = TextStringObject(author)
    if subject is not None:
        retval[NameObject('/Subj')] = TextStringObject(subject)
    if color is not None:
        retval[NameObject('/C')] = floatArray(color)


    return retval


def _popupAnnotation(parent, rect=None):
    '''Create a 'Popup' annotation connected to parent
    (an indirect object).
    '''

    if rect is None:
        # Make Golden ratio rectangle lined up at
        # right-hand side of parent
        _, _, x, y = parent.getObject()['/Rect']
        rect = [x, y-100, x+162, y]

    retval=DictionaryObject({\
            NameObject('/Type'): NameObject('/Annot'),\
          NameObject('/Subtype'): NameObject('/Popup'),\
          NameObject('/M'): now(),\
          NameObject('/Rect'): floatArray(rect),\
          NameObject('/Parent'): parent,\
        })

    return retval


def createHighlight(rect, contents=None, author=None,\
                         subject=None, cdate=None, color=None,\
                         alpha=1, flag=4):
    '''Create a Highlight annotation given rect.

    <rect>: list, coordinates of highlighted rectangles:
            [x0,y0,x1,y1],
            where (x0,y0) is the lower left corner, (x1,y1) the upper
            right corner.

    <contents>: string, content.
    <author>: string, author of the annotation.
    <subject>: string, annotation subject.
    <cdate>: datetime obj or None, creation time. If None, obtain
             the current time.
    <color>: list-type, the color of the highlighted region,
             as an array of type [g], [r,g,b], or [c,m,y,k].
    <alpha>: float, the alpha transparency of the highlight.
    <flag>: int, bit flag of options.  4 means the annotation
            should be printed.  See the PDF spec for more.

    Return: <retval>: A DictionaryObject representing the annotation.

    Update time: 2016-02-19 14:39:13.
    '''

    x0,y0,x1,y1=rect
    qpl=[x0, y1, x1, y1, x0, y0, x1, y0]
    if color is None:
        color=YELLOW
    else:
        try:
            color=getRGBcolor(color)
        except:
            color=YELLOW

    retval = _baseAnno(rect, contents, author,\
            subject, cdate, color, alpha, flag)
    retval[NameObject('/Subtype')] = NameObject('/Highlight')
    retval[NameObject('/QuadPoints')] = floatArray(qpl)

    return retval


def createNote(rect, contents=None, author=None, subject=None,\
                    cdate=None, color=None, alpha=1, flag=4,\
                    icon=None, open_=True, state=None, state_model=None):
    '''Create a text annotation given rect as a sticky note.

    <rect>: a rectangle [x0,y0,x1,y1].  The icon will be in the
            top-left corner of this rectangle (x0,y1) regardless
            of the size.
    <cdate>: datetime obj or None, creation time. If None, obtain
             the current time.
    <color>: list-type, the color of the highlighted region,
             as an array of type [g], [r,g,b], or [c,m,y,k].
    <alpha>: float, the alpha transparency of the highlight.
    <flag>: int, bit flag of options.  4 means the annotation
            should be printed.  See the PDF spec for more.
    <icon>: string, the icon to use for the note.
            Try "Comment", "Key", "Note", "Help", "NewParagraph",
            "Paragraph", "Insert"
    <open_>: bool, whether the note should be opened by default.
    <state>: ?, set the state of the annotation. See the PDF
            spec state_model for further details.

    Return: <retval>: A DictionaryObject representing the annotation.

    Update time: 2016-02-19 14:39:13.
    '''

    if color is None:
        color=YELLOW
    else:
        try:
            color=getRGBcolor(color)
        except:
            color=YELLOW

    retval = _baseAnno(rect, contents, author, subject, cdate,
                                color, alpha, flag)
    retval.popup = True

    retval[NameObject('/Subtype')] = NameObject('/Text')
    retval[NameObject('/Open')] = BooleanObject(open_)

    if icon is not None:
        retval[NameObject('/Name')] = NameObject('/' + icon)
    if state is not None:
        retval[NameObject('/State')] = TextStringObject(state)
    if state_model is not None:
        retval[NameObject('/StateModel')] = TextStringObject(state_model)

    return retval


def addAnnotation(page, outpdf, anno):
    """Add the annotation to the output PDF.

    """
    # We need to make an indirect reference, or Acrobat will get huffy.
    indir = outpdf._addObject(anno)
    if '/Annots' in page:
        page['/Annots'].append(indir)
    else:
        page[NameObject('/Annots')] = ArrayObject([indir])

    if anno.popup:
        popup = _popupAnnotation(indir)
        indir_popup = outpdf._addObject(popup)
        anno[NameObject('/Popup')] =  indir_popup
        page['/Annots'].append(indir_popup)

    return page










'''
Widgets for running tasks in separate threads, feedback display and task
controls.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from queue import Queue
import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import QObject, QThread, Qt, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QDialogButtonBox


LOGGER=logging.getLogger(__name__)



class Worker(QObject):

    worker_jobdone_signal=pyqtSignal(int) # jobid

    def __init__(self, id, func, jobqueue, outqueue):
        super(Worker,self).__init__()
        '''Worker used in separate thread.

        Args:
            id (int): id for the thread/worker.
            func (function): function object to call in the thread.
            jobqueue (Queue): queue object containing input args.
            outqueue (Queue): queue object storing function returns.

        NOTE: args for func is assumed to have the format (jobid,jobargs)
        return of func is assumed to have the format (return_code, jobid,\
                other_results)
        '''

        self.id=id
        self.func=func
        self.jobqueue=jobqueue
        self.outqueue=outqueue
        self.abort=False


    @pyqtSlot()
    def processJob(self):
        while self.jobqueue.qsize():
            QtWidgets.QApplication.processEvents()
            if self.abort:
                return

            args=self.jobqueue.get()
            jobid=args[0]
            try:
                rec=self.func(*args)
                LOGGER.debug('Job %d done. Remaining queue size: %d.'\
                    %(jobid, self.jobqueue.qsize()))
            except:
                rec=(1,jobid,None)
                LOGGER.debug('Job %d failed. Remaining queue size: %d.'\
                    %(jobid, self.jobqueue.qsize()))
            self.jobqueue.task_done()
            self.outqueue.put(rec)
            self.worker_jobdone_signal.emit(jobid)

        return


class Master(QObject):

    all_done_signal=pyqtSignal()
    donejobs_count_signal=pyqtSignal(int) # NO. of finshed jobs

    def __init__(self, func, joblist, max_threads=4, progressbar=None,
            progressbar_style='classic',
            statusbar=None,
            show_message='',
            post_process_func=None,
            post_process_func_args=(),
            post_process_progress=1,
            close_on_finish=True,
            parent=None):
        '''A controller distributing tasks to workers, collecting results and
        sending feedbacks on task status

        Args:
            func (function): function object to call in the thread.
            joblist( list): list of job tasks, in the format:
                [(jobid1, jobargs), (jobid2, jobargs) ... ]

        Kwargs:
            max_threads (int): maximum number of threads to spawn.
            progressbar (QProgressBar or None): progressbar widget to send
                progress feedback. If None, ignore.
            progressbar_style (str): if 'classic', use percentage wise
                progressbar. if 'busy', use busy progressbar.
            statusbar (QStatusBar): status bar widget to show messages. If
                None, ignore.
            show_message (str): message to show in the status bar when running.
            post_process_func (callable): function object to call on each
                return value of <func>, as a post-process.
            post_process_func_args (tuple): additional arguments for the
                <post_process_func>.
            post_process_progress (int): when using 'classic' progressbar,
                how much percentage should the post-process count.
            close_on_finish (bool): If True, clear status bar message and
                hide progressbar after all jobs done.
            parent (QWidget): parent widget.

        '''
        super(self.__class__,self).__init__()

        self.func=func
        self.joblist=joblist
        self.max_threads=max_threads
        self.progressbar=progressbar
        self.progressbar_style=progressbar_style
        self.statusbar=statusbar
        self.show_message=show_message
        self.post_process_func=post_process_func
        self.post_process_func_args=post_process_func_args
        self.post_process_progress=post_process_progress
        self.close_on_finish=close_on_finish
        self.parent=parent

        self.all_done_signal.connect(self.onAllJobsDone)


    def run(self):

        if self.progressbar:
            if self.progressbar_style=='classic':
                if self.post_process_func is None:
                    self.progressbar.setMaximum(len(self.joblist))
                else:
                    self.progressbar.setMaximum(len(self.joblist)+\
                            self.post_process_progress)
                self.progressbar.setValue(0)
            elif self.progressbar_style=='busy':
                self.progressbar.setMaximum(0)
            else:
                raise Exception("Not defined")
            self.progressbar.setVisible(True)
        if self.statusbar and self.show_message:
            self.statusbar.showMessage(self.show_message)

        self.threads=[]
        self.results=[]
        n_threads=min(self.max_threads,len(self.joblist))
        self.finished=0
        self.finished_jobs=[]

        self.jobqueue=Queue()
        self.outqueue=Queue()
        # populate job queue
        for ii,jobii in enumerate(self.joblist):
            self.jobqueue.put(jobii)

        # start worker threads
        for ii in range(n_threads):
            LOGGER.debug('Create thread %s' %ii)

            tii=QThread()
            wii=Worker(ii,self.func,self.jobqueue,self.outqueue)
            self.threads.append((tii,wii)) # need to keep record of both!

            wii.moveToThread(tii)
            wii.worker_jobdone_signal.connect(self.countJobDone)
            tii.started.connect(wii.processJob)
            tii.start()

        return


    @pyqtSlot(int)
    def countJobDone(self,jobid):

        self.finished+=1
        self.finished_jobs.append(jobid)
        self.donejobs_count_signal.emit(self.finished)
        while self.outqueue.qsize():
            resii=self.outqueue.get()
            self.results.append(resii)
        if self.progressbar and self.progressbar_style=='classic':
            self.progressbar.setValue(self.finished)

        LOGGER.debug('finished = %s. NO of results = %d' %(self.finished,
            len(self.results)))

        if self.finished==len(self.joblist):
            self.all_done_signal.emit()

        return


    @pyqtSlot()
    def onAllJobsDone(self):

        for tii,wii in self.threads:
            tii.quit()
            tii.wait()

        if self.post_process_func is not None:
            LOGGER.debug('Call post-process')

            self.results=self.post_process_func(self.results,
                    *self.post_process_func_args)
            if self.progressbar:
                self.progressbar.setValue(self.progressbar.maximum())

        if self.close_on_finish:
            if self.statusbar:
                self.statusbar.clearMessage()
            if self.progressbar:
                self.progressbar.setVisible(False)
        else:
            if self.statusbar:
                self.statusbar.showMessage('Finished.')
            if self.progressbar:
                self.progressbar.setVisible(False)

        return


    @pyqtSlot()
    def abortJobs(self):
        for tii,wii in self.threads:
            wii.abort=True
            tii.quit()
            tii.wait()
        if self.progressbar:
            self.progressbar.setVisible(False)
        if self.statusbar:
            self.statusbar.clearMessage()

        return


class ThreadRunDialog(QtWidgets.QDialog):

    def __init__(self,func, joblist, show_message='', max_threads=3,
            get_results=False, close_on_finish=True,
            progressbar_style='classic',
            post_process_func=None,
            post_process_func_args=(),
            post_process_progress=1,
            parent=None):
        '''A modal dialog shown when calling some long-lasting tasks, with
        message label, progressbar giving feedbacks.

        Args:
            func (callable): function object to call in the thread.
            joblist( list): list of job tasks, in the format:
                [(jobid1, jobargs), (jobid2, jobargs) ... ]

        Kwargs:
            max_threads (int): maximum number of threads to spawn.
            get_results (bool): if True, set the results to a results
                attribute of the dialog.
            close_on_finish (bool): If True, close dialog when all jobs done.
            progressbar_style (str): if 'classic', use percentage wise
                progressbar. if 'busy', use busy progressbar.
            post_process_func (callable): function object to call on each
                return value of <func>, as a post-process.
            post_process_func_args (tuple): additional arguments for the
                <post_process_func>.
            post_process_progress (int): when using 'classic' progressbar,
                how much percentage should the post-process count.
            parent (QWidget): parent widget.
        '''

        super(self.__class__,self).__init__(parent=parent)

        self.func=func
        self.joblist=joblist
        self.show_message=show_message
        self.max_threads=max_threads
        self.get_results=get_results
        self.close_on_finish=close_on_finish
        self.progressbar_style=progressbar_style
        self.post_process_func=post_process_func
        self.post_process_func_args=post_process_func_args
        self.post_process_progress=post_process_progress
        self.parent=parent

        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle('Processing ...')
        self.resize(500,100)

        va=QtWidgets.QVBoxLayout(self)

        self.label=QtWidgets.QLabel(show_message)
        va.addWidget(self.label)

        self.progressbar=QtWidgets.QProgressBar(self)
        self.progressbar.setMaximum(len(joblist))
        self.progressbar.setValue(0)

        self.buttons=QDialogButtonBox(
            QDialogButtonBox.Ok | QDialogButtonBox.Cancel, Qt.Horizontal, self)

        va.addWidget(self.progressbar)
        va.addWidget(self.buttons)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.abortJobs)

        self.master=Master(func,joblist,self.max_threads,self.progressbar,
                self.progressbar_style, None, '',
                self.post_process_func,
                self.post_process_func_args,
                self.post_process_progress,
                self.close_on_finish,
                None)

        self.ok_button=self.buttons.button(QDialogButtonBox.Ok)
        self.ok_button.setEnabled(False)
        self.cancel_button=self.buttons.button(QDialogButtonBox.Cancel)
        self.master.all_done_signal.connect(self.allJobsDone)

    def exec_(self):

        # have to start master first
        self.master.run()
        super(ThreadRunDialog,self).exec_()

        return


    @pyqtSlot()
    def allJobsDone(self):

        self.cancel_button.setEnabled(False)
        self.ok_button.setEnabled(True)
        self.label.setText('Finished')
        if self.close_on_finish:
            self.accept()

        return


    @pyqtSlot()
    def abortJobs(self):
        self.master.abortJobs()
        if self.get_results:
            self.results=self.master.results
        self.reject()

        return


    @pyqtSlot()
    def accept(self):

        LOGGER.debug('get_results = %s' %self.get_results)
        if self.get_results:
            self.results=self.master.results
        super(self.__class__,self).accept()

        return





'''
This part contains methods that load data into various widgets.

loadLibTree() is the major entrance of data loaded from sqlite.
Folders in the library are created in this function, plus 3 preserved system
folders:

    * All (id='-1'): contains all documents. It can not contain sub-folders.
                     You can not add directly new documents in All, new
                     documents can only be added in a normal folder.
    * Needs Review (id='-2'): contains all documents whose 'confirmed' dict
                              value is 'false', indicating an uncertain
                              state in the correctness of meta-data. Needs-
                              review docs are shown in bold font in the doc
                              table. This folder doesn't contain sub-folders,
                              and does not accept new documents.
    * Trash (id='-3'): can contain sub-folders. Folders in trash, or any other
                       folder inside trash are identified by the _trashed_folder_ids
                       property (see _MainFrameProperties.py). As a document
                       can appear in more than one folder, trashing a folder
                       doesn't necessarily trash a doc within, as the doc may
                       still appear in one or more normal folders. Docs that
                       ONLY appear in folders in _trashed_folder_ids are
                       labelled "orphan" and is denoted by the 'deletionPending'==
                       'true' dict value. Orphan docs won't appear in the All
                       folder. Orphan docs are restored by moving into any
                       normal folder, when their 'deletionPending' value is set
                       to 'false'. Orphan docs that don't belong to any folder
                       inside Trash will be put to the Trash folder itself.

System folders are static in that they can't be deleted, renamed or changed in
the order of appearance. They are not saved in the sqlite database file, but
created everytime a library is loaded.

Folders are defined by their name (str) and id (str). Folder info in a library
is stored in the dict self.folder_dict, with the following structure:

    self.folder_dict[folder_id_in_str] = (folder_name, parentid)

Parentid of a folder points to the id of the parent. All top level folders,
including <Needs Review> and <Trash>, have a parentid of '-1'. <All> folder
doesn't have parentid.  Currently there is no limit on the level of folder
nesting.

Valid characters for folder names include alphanumeric characters plus '_' and
'-'. Name confliction is allowed as long as they don't share the same parent.

A <Default> (id='0') folder is created in a newly created library.

After the folder tree creation, the <All> folder is selected.

Upon selecting a folder, documents within, not including those in its
sub-folders, are loaded into the doc table. This is done in loadDocTable(). Ids
of docs in a folder are stored in the dict self.folder_data, with the following
format:

    self.folder_data[folder_id_in_str] = [doc1_id_in_int, doc2_id_in_int, ...]

Meanwhile, each doc stores a list of folder ids that the doc resides in:

    self.meta_dict[docid_in_int]['folders_l'] = [folder1_id_in_int,
                                                 folder2_id_in_int,
                                                 ...
                                                 ]
    forgive me about the int/str type confusion, I haven't got time to fix this.

Upon calling loadDocTable(), a row is selected (if there is any).

Upon selecting a row in doc table, the meta data tab is populated, in
loadMetaTab(). Note texts are loaded in loadNoteTab(), and bibtex string
is loaded in loadBibTab().


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt
from lib import sqlitedb
from lib import bibparse
from lib.tools import getHLine



def addFolder(parent, folderid, folder_dict):
    """Add a new folder item to the folder tree

    Args:
        parent (QTreeWidgetItem or QTreeWidget): parent widget/item onto which
                                                 to add a new folder.
        folderid (str): id of the new folder.
        folder_dict (dict): dict containing folder defs:

            self.folder_dict[folder_id_in_str] = (folder_name, parentid)

    Sub-folders are added by recursive calls of the function.
    """

    foldername,parentid=folder_dict[folderid]
    fitem=QtWidgets.QTreeWidgetItem([foldername,str(folderid)])
    style=QtWidgets.QApplication.style()
    diropen_icon=style.standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
    fitem.setIcon(0,diropen_icon)
    sub_ids=sqlitedb.getChildFolders(folder_dict,folderid)
    if parentid=='-1':
        parent.addTopLevelItem(fitem)
    else:
        parent.addChild(fitem)
    if len(sub_ids)>0:
        for sii in sub_ids:
            addFolder(fitem,sii,folder_dict)

    return


def prepareDocs(meta_dict, docids):
    """Format meta data of docs for display in the doc table

    Args:
        meta_dict (dict): dict containing meta data of all docs in the library,
                          in the format:

                    self.meta_dict[doc_id_in_int] = DocMeta

                          where DocMeta is a dict. See sqlitedb.py for details.
        docids (list): list of doc ids to format.

    Returns: data (list): each element is a list containing 10 fields of a
                          doc to feed into the doc_table QTableView.
    """

    data=[]
    for ii in docids:
        entryii=meta_dict[ii]
        fav_check=QtWidgets.QCheckBox()
        read_check=QtWidgets.QCheckBox()
        fav_check.setChecked(True if entryii['favourite']=='true' else False)
        read_check.setChecked(True if entryii['read']=='true' else False)
        aii=[ii,
            fav_check,
            read_check,
            entryii['has_file'],
            '; '.join(entryii['authors_l']),
            entryii['title'],
            entryii['publication'],
            entryii['year'],
            entryii['added'],
            entryii['confirmed']
            ]

        data.append(aii)

    return data


class MainFrameLoadData:


    #######################################################################
    #                       Load data into widgets                        #
    #######################################################################


    def loadLibTree(self, db, meta_dict, folder_data, folder_dict):
        """Load folders in a library

        Args:
            db (sqlite connection): sqlite connection of the library.
            meta_dict (dict): dict containing meta data of all docs in the library.
            folder_data (dict): dict containing doc ids in a folder.
            folder_dict (dict): dict containing folder info in the library.
        """

        self.db=db
        self.meta_dict=meta_dict
        self.folder_data=folder_data
        self.folder_dict=folder_dict
        #self.inv_folder_dict={v[0]:k for k,v in self.folder_dict.items()}

        style=QtWidgets.QApplication.style()
        diropen_icon=style.standardIcon(QtWidgets.QStyle.SP_DirOpenIcon)
        needsreview_icon=style.standardIcon(QtWidgets.QStyle.SP_MessageBoxInformation)
        trash_icon=style.standardIcon(QtWidgets.QStyle.SP_TrashIcon)

        #-------------Create preserved folders-------------
        self.all_folder=QtWidgets.QTreeWidgetItem(['All','-1'])
        self.all_folder.setIcon(0,diropen_icon)
        self.libtree.addTopLevelItem(self.all_folder)

        self.needsreview_folder=QtWidgets.QTreeWidgetItem(['Needs Review','-2'])
        self.needsreview_folder.setIcon(0,needsreview_icon)
        self.libtree.addTopLevelItem(self.needsreview_folder)

        self.trash_folder=QtWidgets.QTreeWidgetItem(['Trash','-3'])
        self.trash_folder.setIcon(0,trash_icon)
        self.libtree.addTopLevelItem(self.trash_folder)

        self.sys_folders=[self.all_folder,self.needsreview_folder,self.trash_folder]

        #-------------Get all level 1 folders-------------
        folders1=[(vv[0],kk) for kk,vv in self.folder_dict.items() if\
                vv[1]=='-1']
        folders1.sort()

        self.logger.debug('Level 1 folder ids = %s' %folders1)

        #------------------Add separator------------------
        separator=QtWidgets.QTreeWidgetItem([' ',None])
        separator.setFlags(Qt.NoItemFlags)
        self.libtree.addTopLevelItem(separator)
        h_line=getHLine(None)
        self.libtree.setItemWidget(separator,0,h_line)

        #------------Add folders from database------------
        for fnameii,idii in folders1:
            addFolder(self.libtree,idii,self.folder_dict)

        #---------------Add folders in trash---------------
        trashed_folders=[(vv[0],kk) for kk,vv in self.folder_dict.items()\
                if vv[1]=='-3']

        self.logger.debug('Ids of folders in Trash = %s' %trashed_folders)

        for fnameii,idii in trashed_folders:
            addFolder(self.trash_folder,idii,self.folder_dict)

        self.sortFolders()
        self.libtree.setCurrentItem(self.all_folder)

        self.changed_doc_ids=[] # store ids of changed docs
        self.changed_folder_ids=[] # store ids of changed folders

        return


    def loadDocTable(self, folder=None, docids=None, sortidx=None,
            sel_row=None):
        """Load the doc table

        Kwargs:
            folder ((fname_in_str, fid_in_str) or None ): if tuple, load docs
                   within the folder. If None, load the <All> folder.
            docids (list or None): if list, a list of doc ids to load. If None,
                                   load according to the <folder> arg.
            sortidx (int): int in [0,9], index of the column to sort the table
                           with.
            sel_row (int or None): index of the row to select after loading.
                                   If None, don't change selection.
        """

        tablemodel=self.doc_table.model()
        hh=self.doc_table.horizontalHeader()

        self.logger.debug('Load folder = %s. sort indicator section = %s. sort order = %s'\
                %(folder, hh.sortIndicatorSection(), hh.sortIndicatorOrder()))

        #-----------Get list of doc ids to load-----------
        if docids is None:

            # load All folder
            if folder is None or folder[1]=='-1':
                docids=self.meta_dict.keys()

                self.logger.debug('NO. before subtracting orphan docs: %d'\
                        %len(docids))

                docids=list(set(docids).difference(self._orphan_doc_ids))

                self.logger.info('NO. after subtracting orphan docs: %d'\
                        %len(docids))

            # load any other folder
            else:
                docids=self.folder_data[folder[1]]
                self.logger.info('NO. in folder %s = %d' %(folder[1], len(docids)))

        #-------------Format data to table rows-------------
        data=prepareDocs(self.meta_dict,docids)
        tablemodel.arraydata=data

        #--------------------Sort rows--------------------
        if sortidx is not None and sortidx in range(tablemodel.columnCount(None)):
            self.logger.info('sort idx = %s' %sortidx)

            tablemodel.sort(sortidx,Qt.AscendingOrder)

        if len(data)>0:
            self.enableMetaTab()

            #-------------------Select a given row-------------------
            if sel_row is not None:
                current_row=self.doc_table.currentIndex().row()
                docid=self._current_doc

                self.logger.info('Selected row = %s, docid = %s'\
                        %(current_row, docid))

                self.doc_table.selectRow(sel_row)

                if current_row==sel_row:
                    self.logger.info('@@@@@row not change. force sel doc')
                    self.selDoc(self.doc_table.currentIndex(),None)

            self.status_bar.showMessage('%d rows' %len(data))
        else:
            #------------------Clear meta tab------------------
            self.logger.info('No data to be loaded. Clear meta tab.')
            self.removeFolderHighlights()
            self.clearMetaTab()

        #self.doc_table.viewport().update()
        tablemodel.layoutChanged.emit()

        return


    def loadMetaTab(self, docid=None):
        """Load meta data tab of a doc

        Kwargs:
            docid (int or None): if int, the id of the doc to load.
        """

        self.logger.info('docid = %s' %docid)

        if docid is None:
            return

        fields=self._current_meta_dict.keys()
        if fields is None:
            return

        metaii=self.meta_dict[docid]
        def conv(text):
            if isinstance(text,(str)):
                return text
            else:
                return str(text)

        #---------Loop through fields in meta tab---------
        for fii in fields:
            tii=metaii[fii]
            if tii is None:
                self._current_meta_dict[fii].clear()
                continue
            elif fii=='files_l':
                # show only file name
                self.t_meta.delFileField()
                for fjj in tii:
                    self.t_meta.createFileField(fjj)
            else:
                if isinstance(tii,(list,tuple)):
                    tii=u'; '.join(tii)
                self._current_meta_dict[fii].setText(conv(tii))

            if fii in ['authors_l','abstract','tags_l','keywords_l']:
                if self.t_meta.fold_dict[fii]:
                    # fold long texts
                    self._current_meta_dict[fii].foldText()

        return


    def loadBibTab(self, docid=None):
        """Load bibtex tab of a doc

        Kwargs:
            docid (int or None): if int, the id of the doc to load.
        """

        self.logger.info('docid = %s' %docid)
        if docid is None:
            return

        metaii=self.meta_dict[docid]
        omit_keys=self.settings.value('export/bib/omit_fields', [], str)
        if isinstance(omit_keys,str) and omit_keys=='':
            omit_keys=[]

        path_type=self.settings.value('export/bib/path_type',str)
        if path_type=='absolute':
            prefix=self.settings.value('saving/current_lib_folder',str)
        elif path_type=='relative':
            prefix=''

        text=bibparse.metaDictToBib(0,metaii,omit_keys,prefix)[2]
        self.bib_textedit.setText(text)

        return


    def loadNoteTab(self, docid=None):
        """Load note tab of a doc

        Kwargs:
            docid (int or None): if int, the id of the doc to load.
        """

        self.logger.info('docid = %s' %docid)
        if docid is None:
            return

        noteii=self.meta_dict[docid]['notes']

        self.logger.debug('noteii = %s' %noteii)
        self.note_textedit.clear()
        self.note_textedit.setText(noteii)

        return






'''
This part contains some functions dealing with hide/show if widgets, clipboard
copying and clearing data of widgets.


MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QThread


class SettingsThread(QThread):
    '''A separate thread for writing settings'''
    def __init__(self, setting, key, value):
        super(SettingsThread,self).__init__()
        self.setting=setting
        self.key=key
        self.value=value

    def __del__(self):
        self.wait()

    def run(self):
        self.setting.setValue(self.key,self.value)
        return



class MainFrameOtherSlots:


    view_change_sig=pyqtSignal(str,bool)

    #######################################################################
    #                             Other slots                             #
    #######################################################################

    @pyqtSlot()
    def foldFilterButtonClicked(self):
        '''Hide/show the filter list widget

        This is a slot to the clicked signal of the fold button underneath
        the folder tree.

        It is also called when user toggle view changes in the View menu,
        see ViewChangeTriggered() in _MainWindow.py
        '''

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        if self.filter_list.isVisible():
            self.filter_list.setVisible(False)
            self.fold_filter_button.setArrowType(Qt.UpArrow)
            self.view_change_sig.emit('Toggle Filter List',False)
            if 'Toggle Filter List' in show_widgets:
                show_widgets.remove('Toggle Filter List')
        else:
            self.filter_list.setVisible(True)
            self.fold_filter_button.setArrowType(Qt.DownArrow)
            self.view_change_sig.emit('Toggle Filter List',True)
            if 'Toggle Filter List' not in show_widgets:
                show_widgets.append('Toggle Filter List')

        # write change to settings
        self.setting_thread=SettingsThread(self.settings,
                'view/show_widgets', show_widgets)
        self.setting_thread.start()

        return


    @pyqtSlot()
    def foldTabButtonClicked(self):
        '''Hide/show the tab pane

        This is a slot to the clicked signal of the fold button to the right
        of the doc table.

        It is also called when user toggle view changes in the View menu,
        see ViewChangeTriggered() in _MainWindow.py

        Individual tabs in the tab pane can also be toggled. If all tabs
        are toggled off, hide the tab pane as well.
        '''

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        self.logger.debug('Before show_widgets = %s' %show_widgets)

        if self.tab_pane.isVisible():
            self.tab_pane.setVisible(False)
            self.fold_tab_button.setArrowType(Qt.LeftArrow)
            self.view_change_sig.emit('Toggle Tab Pane',False)

            if 'Toggle Tab Pane' in show_widgets:
                show_widgets.remove('Toggle Tab Pane')
        else:
            self.tab_pane.setVisible(True)
            self.fold_tab_button.setArrowType(Qt.RightArrow)
            self.view_change_sig.emit('Toggle Tab Pane',True)

            if 'Toggle Tab Pane' not in show_widgets:
                show_widgets.append('Toggle Tab Pane')

            # check if any tab page still visible
            has_tab=False

            for kk, vv in self.tab_dict.items():
                tabii, tabnameii=vv
                idx=self.tabs.indexOf(tabii)
                if idx!=-1:
                    # -1 idx means not there
                    has_tab=True

            # if no tab page is still visible, remove tab pane as well.
            if not has_tab:
                for kk, vv in self.tab_dict.items():
                    tabii, tabnameii=vv
                    idx=self.tabs.indexOf(tabii)
                    self.tabs.addTab(tabii, tabnameii)
                    self.view_change_sig.emit(kk,True)

                    if kk not in show_widgets:
                        show_widgets.append(kk)

                if 'Toggle Tab Pane' not in show_widgets:
                    show_widgets.append('Toggle Tab Pane')

        self.setting_thread=SettingsThread(self.settings,
                'view/show_widgets', show_widgets)
        self.setting_thread.start()

        self.logger.debug('After show_widgets = %s' %show_widgets)

        return


    def metaTabViewChange(self, view_name='Toggle Tab Pane'):
        '''Hide/show the meta data tab in the tab pane

        It is called when user toggle view changes in the View menu,
        see ViewChangeTriggered() in _MainWindow.py
        '''

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        self.logger.debug('Before show_widgets = %s' %show_widgets)

        has_tab=False
        for kk, vv in self.tab_dict.items():

            tabii, tabnameii=vv
            idx=self.tabs.indexOf(tabii)
            if idx==-1:
                if view_name==kk:
                    tabii.setVisible(True)
                    self.tabs.addTab(tabii, tabnameii)
                    self.view_change_sig.emit(view_name,True)
                    has_tab=True
                    if view_name not in show_widgets:
                        show_widgets.append(view_name)
            else:
                if view_name==kk:
                    tabii.setVisible(False)
                    self.tabs.removeTab(idx)
                    self.view_change_sig.emit(view_name,False)
                    if view_name in show_widgets:
                        show_widgets.remove(view_name)
                else:
                    has_tab=True

        self.logger.debug('has_tab = %s' %has_tab)

        if not has_tab:
            self.tab_pane.setVisible(False)
            self.fold_tab_button.setArrowType(Qt.LeftArrow)
            self.view_change_sig.emit('Toggle Tab Pane',False)
            if 'Toggle Tab Pane' in show_widgets:
                show_widgets.remove('Toggle Tab Pane')
        else:
            self.tab_pane.setVisible(True)
            self.fold_tab_button.setArrowType(Qt.RightArrow)
            self.view_change_sig.emit('Toggle Tab Pane',True)
            if 'Toggle Tab Pane' not in show_widgets:
                show_widgets.append('Toggle Tab Pane')

        self.setting_thread=SettingsThread(self.settings,
                'view/show_widgets', show_widgets)
        self.setting_thread.start()

        self.logger.debug('After show_widgets = %s' %show_widgets)

        return


    def statusbarViewChange(self):
        '''Hide/show the status bar in the main_frame

        It is called when user toggle view changes in the View menu,
        see ViewChangeTriggered() in _MainWindow.py
        '''

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        if self.status_bar.isVisible():
            self.status_bar.setVisible(False)
            if 'Toggle Status bar' in show_widgets:
                show_widgets.remove('Toggle Status bar')
        else:
            self.status_bar.setVisible(True)
            if 'Toggle Status bar' not in show_widgets:
                show_widgets.append('Toggle Status bar')

        self.setting_thread=SettingsThread(self.settings,
                'view/show_widgets', show_widgets)
        self.setting_thread.start()

        return


    @pyqtSlot()
    def clearDuplicateButtonClicked(self):
        '''Hide the duplicate result header frame above the doc table

        This is a slot to the clicked signal of the clear_duplicate_button
        shown in the duplicate_result_frame.

        It is also called in filterTypeCombboxChange(), which is called
        on selecting a folder. Therefore selecting/switching to a folder
        will automatically hide the duplicate_result_frame.
        '''

        if self.duplicate_result_frame.isVisible():
            self.duplicate_result_frame.setVisible(False)

        if not self.doc_table.isVisible():
            self.doc_table.setVisible(True)

            current_folder=self._current_folder
            if current_folder:
                folder,folderid=current_folder

                # TODO: keep a record of previous sortidx?
                if folder=='All' and folderid=='-1':
                    self.loadDocTable(None,sortidx=4,sel_row=0)
                else:
                    self.loadDocTable((folder,folderid),sortidx=4,sel_row=0)
                #self.doc_table.selectRow(0)

        return


    @pyqtSlot()
    def clearSearchResButtonClicked(self):
        '''Hide the search result header frame above the doc table

        This is a slot to the clicked signal of the clear_searchres_button
        shown in the search_res_frame.

        It is also called in filterTypeCombboxChange(), which is called
        on selecting a folder. Therefore selecting/switching to a folder
        will automatically hide the search_res_frame.
        '''

        if self.search_res_frame.isVisible():
            self.search_res_frame.setVisible(False)

        if not self.doc_table.isVisible():
            self.doc_table.setVisible(True)

            current_folder=self._current_folder
            if current_folder:
                folder,folderid=current_folder

                # TODO: keep a record of previous sortidx?
                if folder=='All' and folderid=='-1':
                    self.loadDocTable(None,sortidx=4,sel_row=0)
                else:
                    self.loadDocTable((folder,folderid),sortidx=4,sel_row=0)
                #self.doc_table.selectRow(0)

        return


    @pyqtSlot()
    def copyBibButtonClicked(self):
        '''Copy texts in bibtex tab to clipboard'''

        self.bib_textedit.selectAll()
        self.bib_textedit.copy()

        return


    def clearData(self):
        '''Clear data from meta tab, doc table, folder tree and filter list

        This is called when closing a library.
        '''

        self.clearMetaTab()
        self.doc_table.model().arraydata=[]
        self.doc_table.model().layoutChanged.emit()
        self.libtree.clear()
        self.filter_item_list.clear()

        self.add_button.setEnabled(False)
        self.add_folder_button.setEnabled(False)
        self.duplicate_check_button.setEnabled(False)

        self.logger.info('Data cleared.')

        return



'''
This is the header of the MainFrame class definition. The entire class def
is split into a few files:

* _MainFrame.py             : the __init__(), and creation of various widgets.
* _MainFrameLoadData.py       load data into various widgets. This is major
                              entrance, in particular, its loadLibTree() call
                              loads in data read from sqlite and triggers all
                              subsequent data loading.
* _MainFrameLibTreeSlots.py : actions involving folders, including creation,
                              re-parenting, renaming, trashing, and restoring
                              from trash.
* _MainFrameDocTableSlots.py: actions involving documents, including selection,
                               right clicking menu, deletion, opening.
* _MainFrameToolBarSlots.py : slots to widgets in the tool bar, including adding
                              new doc, new folder, duplicate checking and searching.
* _MainFrameDataSlots.py    : handles data updating, including updates of the
                              in-memory data dictionary from editing the meta
                              data tabs, from DOI querying, adding docs to
                              folders, and saving the in-memory data to sqlite
                              database.
* _MainFrameFilterListSlots.py: actions in response to filterings in the bottom
                                left filtering widget.
* _MainFrameProperties.py: a few getters for easier access to the current states
                           in various widgets.
* _MainFrameMetaTabSlots.py: actions in the meta data tab.
* _MainFrameOtherSlots.py: some functions dealing with hide/show if widgets,
                           clipboard copying and clearing data of widgets.


MeiTing Trunk

An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''


import logging
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, QTimer, pyqtSlot
from PyQt5.QtGui import QIcon, QFont, QBrush, QColor

import _MainFrameLoadData, _MainFrameDataSlots, _MainFrameToolBarSlots,\
        _MainFrameLibTreeSlots, _MainFrameFilterListSlots, _MainFrameDocTableSlots,\
        _MainFrameMetaTabSlots, _MainFrameOtherSlots, _MainFrameProperties
from lib.tools import getMinSizePolicy, getXMinYExpandSizePolicy, \
        getXExpandYMinSizePolicy, getXExpandYExpandSizePolicy, getHLine
from lib.widgets import MyTreeWidget, TableModel,\
        MyHeaderView, MetaTabScroll, CheckDuplicateFrame, NoteTextEdit,\
        SearchResFrame



class MainFrame(QtWidgets.QWidget,_MainFrameLoadData.MainFrameLoadData,
        _MainFrameDataSlots.MainFrameDataSlots,
        _MainFrameToolBarSlots.MainFrameToolBarSlots,
        _MainFrameLibTreeSlots.MainFrameLibTreeSlots,
        _MainFrameFilterListSlots.MainFrameFilterListSlots,
        _MainFrameDocTableSlots.MainFrameDocTableSlots,
        _MainFrameMetaTabSlots.MainFrameMetaTabSlots,
        _MainFrameOtherSlots.MainFrameOtherSlots,
        _MainFrameProperties.MainFrameProperties
        ):
    def __init__(self,settings):
        super(MainFrame,self).__init__()

        self.settings=settings
        self.logger=logging.getLogger(__name__)
        self.initUI()
        self.auto_save_timer=QTimer(self)
        tinter=self.settings.value('saving/auto_save_min', 1, int)*60*1000 # in msc
        self.auto_save_timer.setInterval(tinter)
        self.auto_save_timer.timeout.connect(self.saveToDatabase)


    def initUI(self):

        v_layout0=QtWidgets.QVBoxLayout(self)
        v_layout0.setContentsMargins(2,5,2,1)

        #-------------------Tool bar row-------------------
        h_layout0=QtWidgets.QHBoxLayout()
        v_layout0.addLayout(h_layout0)

        # Add button
        self.add_button=self.createAddMoreButton()
        self.add_folder_button=self.createAddFolderButton()
        self.duplicate_check_button=self.createDuplicateCheckButton()

        self.add_button.setEnabled(False)
        self.add_folder_button.setEnabled(False)
        self.duplicate_check_button.setEnabled(False)

        h_layout0.addWidget(self.add_button)
        h_layout0.addWidget(self.add_folder_button)
        h_layout0.addWidget(self.duplicate_check_button)

        h_layout0.addStretch()

        # seach bar
        self.search_bar=QtWidgets.QLineEdit(self)
        self.search_bar.setPlaceholderText('Type to search')
        self.search_bar.setFixedWidth(300)
        self.search_bar.setSizePolicy(getMinSizePolicy())
        self.search_bar.returnPressed.connect(self.searchBarClicked)

        # search button
        self.search_button=self.createSearchButton()
        self.search_button.setEnabled(False)

        h_layout0.addWidget(self.search_bar)
        h_layout0.addWidget(self.search_button)

        #-------------Add hline below tool bar-------------
        v_layout0.addWidget(getHLine(self))

        #---------------------H split---------------------
        h_split=QtWidgets.QSplitter(Qt.Horizontal)
        h_split.setSizePolicy(getXExpandYExpandSizePolicy())
        v_layout0.addWidget(h_split)

        #---------------------V split---------------------
        v_split=QtWidgets.QSplitter(Qt.Vertical)
        v_la1=QtWidgets.QVBoxLayout()
        v_la1.setContentsMargins(0,0,0,0)
        v_la1.setSpacing(0)
        fr=QtWidgets.QFrame()

        #-------------------Add lib tree-------------------
        self.libtree=self.createLibTree()
        v_la1.addWidget(self.libtree)

        #-----------Add fold filter list button-----------
        self.fold_filter_button=self.createFoldFilterButton()
        v_la1.addWidget(self.fold_filter_button)

        fr.setLayout(v_la1)
        v_split.addWidget(fr)

        #----------------Add filter list----------------
        self.filter_list=self.createFilterList()

        v_split.addWidget(self.filter_list)
        h_split.addWidget(v_split)

        v_split.setSizes([3,1])
        h=v_split.size().height()
        v_split.setSizes([h*0.65,h*0.35])

        #------------Add clear filtering frame------------
        self.clear_filter_frame=self.createClearFilterFrame()

        frame=QtWidgets.QFrame()
        v_la=QtWidgets.QVBoxLayout()
        v_la.addWidget(self.clear_filter_frame)

        #---------Add duplicate check result frame---------
        self.duplicate_result_frame=self.createDuplicateResultFrame()
        v_la.addWidget(self.duplicate_result_frame)

        #-------------Add search result frame-------------
        self.search_res_frame=self.createSearchResultFrame()
        v_la.addWidget(self.search_res_frame)

        #------------------Add doc table------------------
        self.doc_table=self.createDocTable()
        v_la.addWidget(self.doc_table)

        h_layout=QtWidgets.QHBoxLayout()
        h_layout.setContentsMargins(0,0,0,0)
        h_layout.setSpacing(0)
        h_layout.addLayout(v_la)

        #--------------Add fold/unfold button--------------
        self.fold_tab_button=self.createFoldTabButton()
        h_layout.addWidget(self.fold_tab_button)
        frame.setLayout(h_layout)
        frame.setSizePolicy(getXExpandYExpandSizePolicy())
        h_split.addWidget(frame)

        #------------------Add right pane------------------
        self.tab_pane=QtWidgets.QFrame()
        v_la2=QtWidgets.QVBoxLayout()
        self.tab_pane.setLayout(v_la2)

        #-------------Add confirm review frame-------------
        self.confirm_review_frame=self.createConfirmReviewFrame()
        v_la2.addWidget(self.confirm_review_frame)

        #---------------------Add tabs---------------------
        self.tabs=self.createTabs()
        v_la2.addWidget(self.tabs)

        h_split.addWidget(self.tab_pane)

        #------------------Add status bar------------------
        self.status_bar=QtWidgets.QStatusBar()
        #self.status_bar.setFixedHeight(12)
        self.status_bar.setFont(self.settings.value('display/fonts/statusbar',QFont))
        v_layout0.addWidget(self.status_bar)
        self.status_bar.showMessage('Welcome')

        self.progressbar=QtWidgets.QProgressBar(self)
        self.progressbar.setSizePolicy(getXExpandYMinSizePolicy())
        self.progressbar.setMaximum(1)
        self.progressbar.setVisible(False)
        self.status_bar.addPermanentWidget(self.progressbar)

        # search_res_frame created before status bar
        self.search_res_frame.search_done_sig.connect(self.status_bar.clearMessage)

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        if 'Toggle Status bar' not in show_widgets:
            self.status_bar.setVisible(False)

        h_split.setHandleWidth(4)
        w=h_split.size().width()
        h_split.setCollapsible(2,True)
        h_split.setSizes([w*0.15,w*0.6,w*0.25])

        self.logger.info('Main frame UI inited.')

        self.show()


    #######################################################################
    #                           Create widgets                            #
    #######################################################################


    def createAddMoreButton(self):

        button=QtWidgets.QToolButton(self)
        button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

        @pyqtSlot(QtWidgets.QAction)
        def changeDefaultAction(action):
            '''Change the default action'''

            action_data=action.data()
            for aii, (addii, cdii) in add_buttons.items():
                if aii==action_data:
                    addii.setShortcut('Ctrl+n')
                    cdii.setChecked(True)
                    button.setDefaultAction(addii)
                    button.setText('Add')
                    button.setIcon(QIcon.fromTheme('document-new'))
                    self.settings.setValue('import/default_add_action',aii)
                else:
                    addii.setShortcut('')
                    cdii.setChecked(False)

        #-------Add the default action to the button-------
        add_buttons={}
        menu=QtWidgets.QMenu()
        choose_default_menu=QtWidgets.QMenu('Choose Default Action',menu)
        default_act=self.settings.value('import/default_add_action',type=str)

        for aii in ['Add PDF File', 'Add Bibtex File', 'Add RIS File',
                'Add Entry Manually']:

            add_actionii=menu.addAction(aii)
            cd_actionii=choose_default_menu.addAction(aii)
            cd_actionii.setData(aii)
            cd_actionii.setCheckable(True)

            if aii==default_act:
                add_actionii.setShortcut('Ctrl+n')
                cd_actionii.setChecked(True)
            else:
                add_actionii.setShortcut('')
                cd_actionii.setChecked(False)
            add_buttons[aii]=[add_actionii, cd_actionii]

        choose_default_menu.triggered.connect(changeDefaultAction)
        menu.addMenu(choose_default_menu)

        button.setDefaultAction(add_buttons[default_act][0])

        # these has to happen after setDefaultAction()
        button.setText('Add')
        button.setIcon(QIcon.fromTheme('document-new'))
        button.setMenu(menu)
        button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)

        menu.triggered.connect(self.addActionTriggered)

        return button


    def createAddFolderButton(self):

        button=QtWidgets.QToolButton(self)
        button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

        menu=QtWidgets.QMenu()
        self.create_folder_action=menu.addAction('Create Folder')
        self.create_subfolder_action=menu.addAction('Create Sub Folder')

        button.setDefaultAction(self.create_folder_action)

        button.setText('Create Folder')
        button.setIcon(QIcon.fromTheme('folder-new'))
        button.setMenu(menu)
        button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)

        menu.triggered.connect(self.addFolderButtonClicked)

        return button


    def createDuplicateCheckButton(self):

        button=QtWidgets.QToolButton(self)
        button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        button.setText('Check Duplicates')
        button.setIcon(QIcon.fromTheme('scanner'))
        button.clicked.connect(self.checkDuplicateClicked)

        return button


    def createSearchButton(self):

        button=QtWidgets.QToolButton(self)
        menu=QtWidgets.QMenu()

        search_fields=self.settings.value('search/search_fields',[],str)
        if isinstance(search_fields,str) and search_fields=='':
            search_fields=[]

        # add search fields menu
        for fieldii in ['Authors', 'Title', 'Abstract', 'Keywords', 'Tags',
                'Notes', 'Publication']:
            cbii=QtWidgets.QCheckBox(fieldii, menu)
            if fieldii in search_fields:
                cbii.setChecked(True)
            aii=QtWidgets.QWidgetAction(menu)
            cbii.stateChanged.connect(aii.trigger)
            aii.setDefaultWidget(cbii)
            aii.setText(fieldii)
            menu.addAction(aii)

        menu.addSeparator()

        # add desend into subfolders option
        cbii=QtWidgets.QCheckBox('Include sub-folders', menu)
        if self.settings.value('search/desend_folder',bool):
            cbii.setChecked(True)
        aii=QtWidgets.QWidgetAction(menu)
        cbii.stateChanged.connect(aii.trigger)
        aii.setDefaultWidget(cbii)
        aii.setText('Include sub-folders')
        menu.addAction(aii)

        button.setIcon(QIcon.fromTheme('edit-find'))
        button.setMenu(menu)
        button.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)
        button.clicked.connect(self.searchBarClicked)

        return button


    def createLibTree(self):

        libtree=MyTreeWidget(self)

        libtree.customContextMenuRequested.connect(self.libTreeMenu)
        libtree.folder_move_signal.connect(self.changeFolderParent)
        libtree.folder_del_signal.connect(self.trashFolder)
        libtree.itemDoubleClicked.connect(self.renameFolder)
        libtree.add_doc_to_folder_signal.connect(self.addDocToFolder)
        libtree.selectionModel().selectionChanged.connect(self.selFolder)

        return libtree


    def createFoldFilterButton(self):

        button=QtWidgets.QToolButton(self)
        button.setArrowType(Qt.DownArrow)
        button.clicked.connect(self.foldFilterButtonClicked)
        button.setSizePolicy(getXExpandYMinSizePolicy())
        button.setFixedHeight(10)
        #button.setStyleSheet(
                #''''border-radius: 0; border-width: 1px;
                #border-style: solid; border-color:grey''')

        return button


    def createFilterList(self):

        frame=QtWidgets.QFrame()
        scroll=QtWidgets.QScrollArea(self)
        scroll.setWidgetResizable(True)
        v_layout=QtWidgets.QVBoxLayout()
        v_layout.setContentsMargins(0,0,0,0)

        self.filter_type_combbox=QtWidgets.QComboBox(self)
        self.filter_type_combbox.addItem('Filter by authors')
        self.filter_type_combbox.addItem('Filter by keywords')
        self.filter_type_combbox.addItem('Filter by publications')
        self.filter_type_combbox.addItem('Filter by tags')
        self.filter_type_combbox.currentIndexChanged.connect(
                self.filterTypeCombboxChange)
        self.filter_type_combbox.setSizeAdjustPolicy(
                QtWidgets.QComboBox.AdjustToMinimumContentsLength)

        self.filter_item_list=QtWidgets.QListWidget(self)
        self.filter_item_list.itemClicked.connect(self.filterItemClicked)

        v_layout.addWidget(self.filter_type_combbox)
        v_layout.addWidget(self.filter_item_list)

        frame.setLayout(v_layout)
        scroll.setWidget(frame)

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        if 'Toggle Filter List' not in show_widgets:
            scroll.setVisible(False)

        return scroll


    def createDuplicateResultFrame(self):

        frame=CheckDuplicateFrame(self.settings,self)
        frame.tree.currentItemChanged.connect(self.duplicateResultCurrentChange)
        frame.del_doc_from_folder_signal.connect(self.delFromFolder)
        frame.del_doc_from_lib_signal.connect(self.delDoc)
        frame.clear_duplicate_button.clicked.connect(
                self.clearDuplicateButtonClicked)
        frame.setVisible(False)

        return frame


    def createSearchResultFrame(self):

        frame=SearchResFrame(self.settings,self)
        frame.tree.currentItemChanged.connect(self.searchResultCurrentChange)
        frame.clear_searchres_button.clicked.connect(self.clearSearchResButtonClicked)
        frame.create_folder_sig.connect(self.createFolderFromSearch)
        #frame.hide_doc_sig.connect(self.hideDocTable)
        frame.setVisible(False)

        return frame


    def createClearFilterFrame(self):

        frame=QtWidgets.QFrame()
        frame.setStyleSheet('background: rgb(235,225,190)')
        h_la=QtWidgets.QHBoxLayout()

        # clear fitlering button
        self.clear_filter_button=QtWidgets.QToolButton(self)
        self.clear_filter_button.setText('Clear')
        self.clear_filter_button.clicked.connect(self.clearFilterButtonClicked)

        self.clear_filter_label=QtWidgets.QLabel('  Clear current filtering')
        h_la.addWidget(self.clear_filter_label)
        h_la.addWidget(self.clear_filter_button)

        frame.setLayout(h_la)
        frame.setVisible(False)

        return frame


    def createConfirmReviewFrame(self):

        frame=QtWidgets.QFrame()
        frame.setStyleSheet('background: rgb(235,225,190)')
        h_la=QtWidgets.QHBoxLayout()

        # confirm button
        self.confirm_review_button=QtWidgets.QToolButton(self)
        self.confirm_review_button.setText('Confirm')
        self.confirm_review_button.clicked.connect(self.confirmReviewButtonClicked)

        label=QtWidgets.QLabel('Meta data is correct?')
        h_la.addWidget(label)
        h_la.addWidget(self.confirm_review_button)

        frame.setLayout(h_la)
        frame.setVisible(False)

        return frame


    def createDocTable(self):

        tv=QtWidgets.QTableView(self)

        hh=MyHeaderView(self)

        tv.setHorizontalHeader(hh)
        tv.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        tv.setShowGrid(True)
        tv.setSortingEnabled(True)

        tv.setDragEnabled(True)
        #tv.setSectionsMovable(True)
        tv.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop)

        header=['docid','favourite','read','has_file','author','title',
                'journal','year','added','confirmed']
        tablemodel=TableModel(self,[],header,self.settings)
        tablemodel.dataChanged.connect(self.modelDataChanged)
        tv.setModel(tablemodel)
        hh.setModel(tablemodel)
        hh.initresizeSections()
        tv.setColumnHidden(0,True) # doc id column, hide
        tv.setColumnHidden(9,True) # needs review column, shown as bold/normal

        tv.selectionModel().currentChanged.connect(self.selDoc)
        tv.clicked.connect(self.docTableClicked)
        #tablemodel.rowsInserted.connect(self.model_insert_row)
        tv.setContextMenuPolicy(Qt.CustomContextMenu)
        tv.customContextMenuRequested.connect(self.docTableMenu)

        tv.doubleClicked.connect(self.docDoubleClicked)
        tv.setAlternatingRowColors(True)

        tv.setStyleSheet('''alternate-background-color: rgb(230,230,249);
                background-color: none''')

        return tv


    def createFoldTabButton(self):

        button=QtWidgets.QToolButton(self)
        button.setArrowType(Qt.RightArrow)
        button.clicked.connect(self.foldTabButtonClicked)
        button.setFixedWidth(10)
        button.setFixedHeight(200)
        #button.setStyleSheet(
                #''''border-radius: 0; border-width: 1px;
                #border-style: solid; border-color:grey''')

        return button


    def createTabs(self):

        tabs=QtWidgets.QTabWidget()

        self.t_notes=self.createNoteTab()
        self.t_bib=self.createBiBTab()
        self.t_scratchpad=self.createScratchTab()
        self.t_meta=MetaTabScroll(self.settings,self)
        self.t_meta.meta_edited.connect(lambda field_list: self.updateTableData(\
            self._current_doc,self.t_meta._meta_dict,field_list))
        self.t_meta.update_by_doi_signal.connect(self.updateByDOI)

        # tab_dict is used for visibility control
        self.tab_dict={'Toggle Meta Tab': [self.t_meta, 'Meta Data'],
                'Toggle Notes Tab': [self.t_notes, 'Notes'],
                'Toggle BibTex Tab': [self.t_bib, 'BibTex'],
                'Toggle Scratch Pad Tab': [self.t_scratchpad, 'Scratch Pad']
                }

        show_widgets=self.settings.value('view/show_widgets',[],str)
        if isinstance(show_widgets,str) and show_widgets=='':
            show_widgets=[]

        # add tabs that are toggled on in view menu
        for tii in show_widgets:
            if tii in self.tab_dict:
                tabs.addTab(*self.tab_dict[tii])

        for tii in list(set(self.tab_dict.keys()).difference(show_widgets)):
            self.tab_dict[tii][0].setVisible(False)

        return tabs


    def createNoteTab(self):

        scroll=QtWidgets.QScrollArea(self)
        scroll.setWidgetResizable(True)
        frame=QtWidgets.QFrame()
        v_layout=QtWidgets.QVBoxLayout()

        self.note_textedit=NoteTextEdit(self.settings)
        self.note_textedit.note_edited_signal.connect(lambda: self.updateNotes(
            self._current_doc,self.note_textedit.toPlainText()))

        v_layout.addWidget(self.note_textedit)
        frame.setLayout(v_layout)
        scroll.setWidget(frame)

        return scroll


    def createScratchTab(self):

        scroll=QtWidgets.QScrollArea(self)
        scroll.setWidgetResizable(True)
        frame=QtWidgets.QFrame()
        v_layout=QtWidgets.QVBoxLayout()

        self.scratchpad_textedit=QtWidgets.QTextEdit(self)
        self.scratchpad_textedit.setFont(self.settings.value(
            '/display/fonts/scratch_pad',QFont))
        self.scratchpad_textedit.setSizePolicy(getXExpandYExpandSizePolicy())

        v_layout.addWidget(self.scratchpad_textedit)
        frame.setLayout(v_layout)
        scroll.setWidget(frame)

        return scroll


    def createBiBTab(self):

        frame=QtWidgets.QWidget()
        scroll=QtWidgets.QScrollArea(self)
        scroll.setWidgetResizable(True)
        scroll.setWidget(frame)
        v_layout=QtWidgets.QVBoxLayout()

        self.copy_bib_button=QtWidgets.QToolButton(self)
        self.copy_bib_button.setText('Copy')
        self.copy_bib_button.setIcon(QIcon.fromTheme('edit-copy'))
        self.copy_bib_button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)

        h_layout=QtWidgets.QHBoxLayout()
        h_layout.addWidget(self.copy_bib_button)
        h_layout.addStretch()

        self.bib_textedit=QtWidgets.QTextEdit(self)
        self.bib_textedit.setReadOnly(True)
        self.bib_textedit.setFont(self.settings.value('/display/fonts/bibtex',QFont))
        v_layout.addLayout(h_layout)
        v_layout.addWidget(self.bib_textedit)
        frame.setLayout(v_layout)
        self.copy_bib_button.clicked.connect(self.copyBibButtonClicked)

        return scroll




'''
Export dialog.

MeiTing Trunk
An open source reference management tool developed in PyQt5 and Python3.

Copyright 2018-2019 Guang-zhi XU

This file is distributed under the terms of the
GPLv3 licence. See the LICENSE file for details.
You may use, distribute and modify this code under the
terms of the GPLv3 license.
'''

import os
import shutil
import logging
from collections import OrderedDict
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt,\
        pyqtSignal,\
        pyqtSlot
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QDialogButtonBox
from .. import sqlitedb
from .. import bibparse
from .. import risparse
from ..tools import getHLine, createFolderTree, iterTreeWidgetItems, autoRename
from .threadrun_dialog import ThreadRunDialog
from .fail_dialog import FailDialog

LOGGER=logging.getLogger(__name__)



class ExportDialog(QtWidgets.QDialog):
    def __init__(self,settings,parent):
        '''
        Args:
            parent (QWidget): parent widget.
            settings (QSettings): application settings. See _MainWindow.py
        '''

        super(ExportDialog,self).__init__(parent=parent)

        self.settings=settings
        self.parent=parent

        self.label_color='color: rgb(0,0,140); background-color: rgb(235,235,240)'
        self.title_label_font=QFont('Serif',12,QFont.Bold)
        self.sub_title_label_font=QFont('Serif',10,QFont.Bold)

        self.resize(900,600)
        self.setWindowTitle('Bulk Export')
        self.setWindowModality(Qt.ApplicationModal)

        v_layout=QtWidgets.QVBoxLayout()
        h_layout=QtWidgets.QHBoxLayout()
        #h_layout.setContentsMargins(10,40,10,20)
        self.setLayout(v_layout)

        title_label=QtWidgets.QLabel('    Choose Export Type')
        title_label.setFont(QFont('Serif',12,QFont.Bold))
        v_layout.addWidget(title_label)

        v_layout.addLayout(h_layout)

        self.cate_list=QtWidgets.QListWidget(self)
        #self.list.setSizePolicy(getXMinYExpandSizePolicy())
        self.cate_list.setMaximumWidth(200)
        h_layout.addWidget(self.cate_list)

        #self.cate_list.setStyleSheet('''
            #QListWidget::item { border: 0px solid rgb(235,235,240);
            #font: 14px;
            #background-color: rgb(205,205,245);
            #color: rgb(100,10,13) };
            #background-color: rgb(230,234,235);
            #''')

        self.cate_list.addItems(['Copy Document Files', 'Export to bibtex',
            'Export to RIS', 'Export to Zotero'])

        self.content_vlayout=QtWidgets.QVBoxLayout()
        h_layout.addLayout(self.content_vlayout)

        if self.parent.is_loaded:
            # What if database if empty?
            folder_dict=self.parent.main_frame.folder_dict
            self.folder_tree=createFolderTree(folder_dict,self)
        else:
            self.folder_tree=None

        self.buttons=QDialogButtonBox(QDialogButtonBox.Close,
            Qt.Horizontal, self)
        self.export_button=self.buttons.addButton('Export',
                QDialogButtonBox.ApplyRole)

        self.export_button.clicked.connect(self.doExport)
        self.buttons.rejected.connect(self.reject)

        self.content_vlayout.addWidget(self.buttons)

        self.cate_list.currentItemChanged.connect(self.cateSelected)
        self.cate_list.setCurrentRow(0)



    @pyqtSlot(QtWidgets.QListWidgetItem)
    def cateSelected(self,item):
        '''Load widgets for a selected category

        Args:
            item (QListWidgetItem): selected category item.
        '''

        item_text=item.text()
        LOGGER.debug('item.text() = %s' %item_text)

        if self.content_vlayout.count()>1:
            self.content_vlayout.removeWidget(self.content_frame)

        if item_text=='Copy Document Files':
            self.content_frame=self.loadCopyFileOptions()
        elif item_text=='Export to bibtex':
            self.content_frame=self.loadExportBibOptions()
        elif item_text=='Export to RIS':
            self.content_frame=self.loadExportRISOptions()
        elif item_text=='Export to Zotero':
            self.content_frame=self.loadExportZoteroOptions()

        self.content_vlayout.insertWidget(0,self.content_frame)

        return


    def createFrame(self,title):
        '''Create a template frame for a category page

        Args:
            title (str): title of the category

        Returns:
            scroll (QScrollArea): a scroll area.
            va (QVBoxLayout): the vertical box layout used in scroll.
        '''

        frame=QtWidgets.QWidget(self)
        scroll=QtWidgets.QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setWidget(frame)
        va=QtWidgets.QVBoxLayout()
        frame.setLayout(va)
        va.setSpacing(int(va.spacing()*2))

        label=QtWidgets.QLabel(title)
        label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)
        va.addWidget(label)
        #va.addWidget(getHLine(self))

        return scroll, va


    def clearFolderTreeState(self):
        '''Uncheck all checkboxes in the folder tree'''

        for item in iterTreeWidgetItems(self.folder_tree):
            item.setCheckState(0,False)

        return


    def getFolderTreeState(self):
        '''Collect check states in the folder tree

        Returns:
            folders (list): list of QTreeWidgetItem who is checked.
        '''

        folders=[]
        for item in iterTreeWidgetItems(self.folder_tree):
            if item.data(1,0)=='-1':
                # all folder
                continue
            if item.checkState(0):
                folders.append(item)

        return folders


    def createOmitKeyGroup(self):
        '''Create checkbox group
        '''

        grid=QtWidgets.QGridLayout()

        self.groupbox=QtWidgets.QGroupBox('Omit these fields in the exported bib entry.')
        self.groupbox.setCheckable(True)

        omittable_keys=[
            'read', 'favourite', 'added', 'confirmed', 'firstNames_l',
            'lastName_l', 'deletionPending', 'folders_l', 'type', 'id',
            'abstract', 'advisor', 'month', 'language', 'confirmed',
            'deletionPending', 'note', 'publicLawNumber', 'sections',
            'reviewedArticle', 'userType', 'shortTitle', 'sourceType',
            'code', 'codeNumber', 'codeSection', 'codeVolume', 'citationKey',
            'day', 'dateAccessed', 'internationalAuthor', 'internationalUserType',
            'internationalTitle', 'internationalNumber', 'genre', 'lastUpdate',
            'legalStatus', 'length', 'medium'
            ]
        omittable_keys.sort()

        omit_keys=self.settings.value('export/bib/omit_fields', [], str)
        # likely something wrong with qt. When list is set empty by
        # settings.setValue('key',[]), on the NEXT load of the program,
        # settings.value('export/bib/omit_fields', [], str) gives ''
        if isinstance(omit_keys,str) and omit_keys=='':
            omit_keys=[]

        for ii,keyii in enumerate(omittable_keys):
            checkboxii=QtWidgets.QCheckBox(keyii,self.groupbox)
            if keyii in omit_keys:
                checkboxii.setChecked(True)
            checkboxii.stateChanged.connect(self.omitKeyChanged)
            grid.addWidget(checkboxii,int(ii/3),ii%3)

        self.groupbox.toggled.connect(lambda on: self.omitKeysGroupChanged(on,
            self.groupbox))

        self.groupbox.setLayout(grid)

        return self.groupbox


    def loadCopyFileOptions(self):
        '''Load widgets for the Copy Files category'''

        scroll,va=self.createFrame('Copy Document Files')
        self.current_task='copy_file'

        #---------------Folder choice section---------------
        label=QtWidgets.QLabel('''
        Choose folders to export documents. <br/>
        This will copy documents (e.g. PDFs) from the
        <span style="font:bold;">"_collections"</span>
        folder to a separate folder under <span style="font:bold;">"%s/"</span>
        ''' %self.settings.value('saving/current_lib_folder',str))
        label.setTextFormat(Qt.RichText)
        label.setWordWrap(True)
        va.addWidget(label)

        if self.folder_tree:
            self.clearFolderTreeState()
            va.addWidget(self.folder_tree)
        else:
            va.addWidget(QtWidgets.QLabel('Library empty'))
            self.export_button.setEnabled(False)

        return scroll


    def loadExportBibOptions(self):
        '''Load widgets for the Export Bibtex category'''

        scroll,va=self.createFrame('Export to bibtex')
        self.bib_settings={}

        self.current_task='bib_export'

        #--------------Folder choice section--------------
        label=QtWidgets.QLabel('Choose folder(s) to export.')
        va.addWidget(label)

        if self.folder_tree:
            self.folder_tree.setMinimumHeight(300)
            self.clearFolderTreeState()
            va.addWidget(self.folder_tree)
        else:
            va.addWidget(QtWidgets.QLabel('Library empty'))
            self.export_button.setEnabled(False)

        #--------------Export manner section--------------
        self.radio_groupbox=QtWidgets.QGroupBox('Saving manner')
        ha2=QtWidgets.QHBoxLayout()
        self.radio_groupbox.setLayout(ha2)

        choices=['All in one', 'Per folder', 'Per document']
        for ii in choices:
            radioii=QtWidgets.QRadioButton(ii)
            if ii=='Per folder':
                radioii.setChecked(True)
            ha2.addWidget(radioii)

        va.addWidget(self.radio_groupbox)

        #----------------Omit keys section----------------
        self.omitkey_groupbox=self.createOmitKeyGroup()
        va.addWidget(self.omitkey_groupbox)

        #--------------File path type section--------------
        self.path_type_groupbox=self.createPathTypeGroup('bib')
        va.addWidget(self.path_type_groupbox)

        return scroll


    def createPathTypeGroup(self, export_type):
        '''Create radiobutton group for file path type selection'''

        groupbox=QtWidgets.QGroupBox('Use relative or absolute paths for files.')
        ha=QtWidgets.QHBoxLayout()
        path_type=self.settings.value('export/%s/path_type' %export_type,str)

        for ii in ['Relative', 'Absolute']:
            buttonii=QtWidgets.QRadioButton(ii)
            if ii.lower()==path_type:
                buttonii.setChecked(True)
            ha.addWidget(buttonii)

        groupbox.setLayout(ha)

        return groupbox


    def getPathType(self):
        '''Collect check states in the path type radiobutton group'''

        for rii in self.path_type_groupbox.findChildren(QtWidgets.QRadioButton):
            if rii.isChecked():
                return rii.text().lower()


    def getExportManner(self):
        '''Collect check states in the export manner radiobutton group'''

        for box in self.radio_groupbox.findChildren(QtWidgets.QRadioButton):
            if box.isChecked():
                choice=box.text()
                break

        return choice


    def omitKeyChanged(self,on):
        '''Store changes in the omit keys checkbox group'''

        self.bib_settings['omit_keys']=self.getOmitKeys()
        LOGGER.debug('omit keys = %s' %self.bib_settings['omit_keys'])

        return


    def omitKeysGroupChanged(self, on, groupbox):
        '''Change check states in the omit keys checkbox group as a whole'''

        omit_keys=[]

        for box in groupbox.findChildren(QtWidgets.QCheckBox):
            box.stateChanged.disconnect()
            box.setChecked(on)
            box.setEnabled(True)
            box.stateChanged.connect(self.omitKeyChanged)
            if box.isChecked():
                omit_keys.append(box.text())

        return


    def getOmitKeys(self):
        '''Collect check states in the omit key checkbox group'''

        omit_keys=[]

        for box in self.omitkey_groupbox.findChildren(QtWidgets.QCheckBox):
            if box.isChecked():
                omit_keys.append(box.text())

        return omit_keys


    def loadExportRISOptions(self):
        '''Load widgets for the Export RIS category'''

        scroll,va=self.createFrame('Export to RIS')
        self.current_task='ris_export'

        #self.ris_settings={}

        #--------------Folder choice section--------------
        label=QtWidgets.QLabel('Choose folder(s) to export.')
        va.addWidget(label)

        if self.folder_tree:
            self.folder_tree.setMinimumHeight(300)
            self.clearFolderTreeState()
            va.addWidget(self.folder_tree)
        else:
            va.addWidget(QtWidgets.QLabel('Library empty'))
            self.export_button.setEnabled(False)

        #--------------Export manner section--------------
        self.radio_groupbox=QtWidgets.QGroupBox('Saving manner')
        ha2=QtWidgets.QHBoxLayout()
        self.radio_groupbox.setLayout(ha2)

        choices=['All in one', 'Per folder', 'Per document']
        for ii in choices:
            radioii=QtWidgets.QRadioButton(ii)
            if ii=='Per folder':
                radioii.setChecked(True)
            ha2.addWidget(radioii)

        va.addWidget(self.radio_groupbox)

        #--------------File path type section--------------
        self.path_type_groupbox=self.createPathTypeGroup('bib')
        va.addWidget(self.path_type_groupbox)

        return scroll


    def loadExportZoteroOptions(self):
        '''Load widgets for the Export Zotero category'''

        scroll,va=self.createFrame('Export to Zotero')
        self.current_task='zotero_export'

        return scroll


    def doExportFiles(self):
        '''Export files'''

        folder_dict=self.parent.main_frame.folder_dict
        folder_data=self.parent.main_frame.folder_data
        meta_dict=self.parent.main_frame.meta_dict
        #storage_folder=self.settings.value('saving/storage_folder',str)
        lib_folder=self.settings.value('saving/current_lib_folder',str)

        if not os.path.exists(lib_folder):
            LOGGER.warning('Cant find storage folder (%s) when exporting files.'\
                    %lib_folder)

            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Critical)
            msg.setWindowTitle('Critical Error!')
            msg.setText("Can not find storage folder.")
            msg.setInformativeText("I can't find the storage folder at %s. Data lost."\
                    %lib_folder)
            msg.exec_()
            return

        folders=self.getFolderTreeState()
        if len(folders)==0:
            self.popUpChooseFolder()
            return

        LOGGER.debug('Current lib folder = %s' %lib_folder)

        job_list=[] # (jobid, source_path, target_path)
        for item in folders:
            folderid=item.data(1,0)
            docids=folder_data[folderid]
            tree=sqlitedb.getFolderTree(folder_dict,folderid)[1]

            LOGGER.debug('Process folder %s, id = %s, tree = %s'\
                    %(item.data(0,0), item.data(1,0), tree))

            folderii=os.path.join(lib_folder,tree)
            if not os.path.exists(folderii):
                os.makedirs(folderii)
                LOGGER.info('Create folder %s' %folderii)

            for idii in docids:
                if meta_dict[idii]['has_file']:
                    for fjj in meta_dict[idii]['files_l']:
                        filenamejj=sqlitedb.renameFile(fjj,meta_dict[idii])
                        newfjj=os.path.join(folderii,filenamejj)
                        # prepend lib_folder to rel file path
                        oldjj=os.path.join(lib_folder,fjj)
                        job_list.append((len(job_list), oldjj, newfjj))

        def copyFunc(jobid,s,t):
            try:
                shutil.copy(s,t)
                LOGGER.debug('Copied file %s to %s' %(s,t))
                rec=0
                result=None
            except:
                LOGGER.exception('Failed to copy %s to %s' %(s,t))
                rec=1
                result=None

            return rec,jobid,result

        if len(job_list)>0:
            thread_run_dialog=ThreadRunDialog(copyFunc,job_list,
                    show_message='Exporting Files...',max_threads=4,
                    get_results=False,
                    close_on_finish=False,
                    progressbar_style='classic',parent=self)

            thread_run_dialog.exec_()

        return


    def doBibExport(self):

        #---------------Get selected folders---------------
        folders=[(item.data(0,0), item.data(1,0)) for item in self.getFolderTreeState()]
        if len(folders)==0:
            self.popUpChooseFolder()
            return

        #------------Popup for saving location------------
        manner=self.getExportManner()
        if manner=='All in one':
            default_path='bibtex_export.bib'
            fname = QtWidgets.QFileDialog.getSaveFileName(self,
                    'Save Citaitons to bib File',
                    default_path,
                    "bib Files (*.bib);; All files (*)")[0]
        else:
            fname=QtWidgets.QFileDialog.getExistingDirectory(self,
                'Choose a folder to save bibtex files')

        if not fname:
            return

        #------------------Get path type------------------
        path_type=self.getPathType()
        if path_type=='absolute':
            prefix=self.settings.value('saving/current_lib_folder',str)
        elif path_type=='relative':
            prefix=''

        LOGGER.info('Selected folders = %s' %folders)
        LOGGER.info('Saving manner = %s' %manner)
        LOGGER.info('Chosen bib file = %s' %fname)
        LOGGER.info('path_prefix = %s' %prefix)

        #-----------------Prepare job list-----------------
        folder_data=self.parent.main_frame.folder_data
        meta_dict=self.parent.main_frame.meta_dict
        omit_keys=self.getOmitKeys()

        LOGGER.debug('omit keys = %s' %omit_keys)

        #-----------------Create job list-----------------
        job_list=[]
        citationkeys=[]
        docs=[]

        if manner in ['All in one', 'Per document']:

            for folderii in folders:
                docs.extend(folder_data[folderii[1]])

            docs=list(set(docs))
            n=0

            for docii in docs:
                cii=meta_dict[docii]['citationkey']
                if cii=='':
                    cii=str(docii)
                citationkeys.append(cii)
                job_list.append((n, meta_dict[docii], omit_keys, prefix))
                n+=1

            #---------------Sort by citationkey---------------
            if manner=='All in one':
                job_list=[x for _,x in sorted(zip(citationkeys,job_list))]
                citationkeys=sorted(citationkeys)

        elif manner=='Per folder':

            n=0
            for foldernameii,fidii in folders:

                jobsii=[]
                #----------------Loop through docs----------------
                docids=folder_data[fidii]
                citationkeysii=[]
                for docii in docids:
                    if docii in docs:
                        # remove duplicates
                        continue

                    cii=meta_dict[docii]['citationkey']
                    if cii=='':
                        cii=str(docii)
                    citationkeysii.append(cii)
                    jobsii.append((n, meta_dict[docii], omit_keys, prefix))
                    n+=1
                    docs.append(docii)

                #---------------Sort by citationkey---------------
                jobsii=[x for _,x in sorted(zip(citationkeysii,jobsii))]
                citationkeysii=sorted(citationkeysii)
                job_list.extend(jobsii)
                citationkeys.extend(citationkeysii)

        #------------------Run in thread------------------
        thread_run_dialog=ThreadRunDialog(bibparse.metaDictToBib, job_list,
                show_message='Processing...', max_threads=1, get_results=False,
                close_on_finish=False,
                progressbar_style='classic',
                post_process_func=self.saveBib,
                post_process_func_args=(folders, manner, fname, folder_data,
                    meta_dict,citationkeys),
                parent=self)

        thread_run_dialog.exec_()

        return


    def saveBib(self, results, folders, manner, fname, folder_data, meta_dict,
            citationkeys):
        """Save exported bibtex text to files

        Args:
            results (list): return values of metaDictToBib().
            folders (list): list of folders in the format [(name, folderid), ].
            manner (str): saving manner, 'All in one': save all bib entries
                          in a single file.
                          'Per folder': group by folder.
                          'Per document': per doc.
            fname (str): if manner=='All in one', the abspath to the output
                         bib file. Otherwise, the abspath to folder to save
                         output bib files.
            folder_data (dict): documents in each folder. keys: folder id in str,
                                values: list of doc ids. To determine when
                                a folder gets all its docs exported.
            meta_dict (dict): meta data of all documents. keys: docid,
                              values: DocMeta dict.
            citationkeys (list): list of the citationkeys for each doc. Used
                                 as the file name in 'Per document' mode.

        """

        faillist=[]

        if manner=='All in one':
            text=''
            for recii,jobii,textii,docii in results:
                if recii==0:
                    text=text+textii+'\n'
                elif recii==1:
                    faillist.append(docii)

            with open(fname,'w') as fout:
                fout.write(text)

            LOGGER.info('Saved combined outputs to file %s' %fname)

        elif manner=='Per document':
            for recii,jobii,textii,docii in results:
                if recii==0:
                    citationkey=citationkeys[jobii]
                    fnameii='%s.bib' %citationkey
                    fnameii=os.path.join(fname,fnameii)
                    fnameii=autoRename(fnameii)

                    with open(fnameii,'w') as fout:
                        fout.write(textii)

                    LOGGER.info('Saved per doc output to file %s' %fnameii)

                elif recii==1:
                    faillist.append(docii)

        elif manner=='Per folder':

            #-------------Prepare result counting-------------
            folder_counts={}
            folder_results={}
            for foldernameii,fidii in folders:
                folder_results[fidii]=[]
                folder_counts[fidii]=0

            for recii,jobii,textii,docii in results:
                if recii==1:
                    faillist.append(docii)

                for fnamejj,fidjj in folders:
                    if docii in folder_data[fidjj]:
                        folder_counts[fidjj]+=1
                        if recii==0:
                            folder_results[fidjj].append(textii)

                        if folder_counts[fidjj]==len(folder_data[fidjj]):
                            LOGGER.debug('folder %s got all data. Save.' %fnamejj)

                            fnamejj=os.path.join(fname,'%s.bib' %fnamejj)
                            text='\n'.join(folder_results[fidjj])

                            with open(fnamejj,'w') as fout:
                                fout.write(text)

                            LOGGER.info('Saved per folder output to file %s' %fnamejj)

        #-----------------Show failed jobs-----------------
        if len(faillist)>0:

            fail_entries=[]
            for docii in faillist:
                metaii=meta_dict[docii]
                entryii='* %s_%s_%s' %(', '.join(metaii['authors_l']),
                        metaii['year'],
                        metaii['title'])

                LOGGER.warning('Failed export job: %s' %entryii)

                fail_entries.append(entryii)

            '''
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setWindowTitle('Error')
            msg.setText('Oopsie.')
            msg.setInformativeText('Failed to export some entries')
            msg.setDetailedText('\n'.join(fail_entries))
            msg.exec_()
            '''
            msg=FailDialog()
            msg.setText('Oopsie.')
            msg.setInformativeText('Failed to export some entries.')
            msg.setDetailedText('\n'.join(fail_entries))
            msg.create_fail_summary.connect(lambda:\
                    self.parent.main_frame.createFailFolder('bibtex export',
                        faillist))
            msg.exec_()

        return


    def doRISExport(self):

        #---------------Get selected folders---------------
        folders=[(item.data(0,0), item.data(1,0)) for item in self.getFolderTreeState()]
        if len(folders)==0:
            self.popUpChooseFolder()
            return

        #------------Popup for saving location------------
        manner=self.getExportManner()
        if manner=='All in one':
            default_path='RIS_export.ris'
            fname = QtWidgets.QFileDialog.getSaveFileName(self,
                    'Save Citaitons to RIS File',
                    default_path,
                    "ris Files (*.ris);; All files (*)")[0]
        else:
            fname=QtWidgets.QFileDialog.getExistingDirectory(self,
                'Choose a folder to save RIS files')

        if not fname:
            return

        #------------------Get path type------------------
        path_type=self.getPathType()
        if path_type=='absolute':
            prefix=self.settings.value('saving/current_lib_folder',str)
        elif path_type=='relative':
            prefix=''

        LOGGER.info('Selected folders = %s' %folders)
        LOGGER.info('Saving manner = %s' %manner)
        LOGGER.info('Chosen ris file = %s' %fname)
        LOGGER.info('path_prefix = %s' %prefix)

        #-----------------Prepare job list-----------------
        folder_data=self.parent.main_frame.folder_data
        meta_dict=self.parent.main_frame.meta_dict

        #-----------------Create job list-----------------
        job_list=[]
        citationkeys=[]
        docs=[]

        if manner in ['All in one', 'Per document']:

            for folderii in folders:
                docs.extend(folder_data[folderii[1]])

            docs=list(set(docs))
            n=0
            for docii in docs:
                cii=meta_dict[docii]['citationkey']
                if cii=='':
                    cii=str(docii)
                citationkeys.append(cii)
                job_list.append((n, meta_dict[docii], prefix))
                n+=1

            #---------------Sort by citationkey---------------
            if manner=='All in one':
                job_list=[x for _,x in sorted(zip(citationkeys,job_list))]
                citationkeys=sorted(citationkeys)

        elif manner=='Per folder':

            n=0
            for foldernameii,fidii in folders:

                jobsii=[]
                #----------------Loop through docs----------------
                docids=folder_data[fidii]
                citationkeysii=[]
                for docii in docids:
                    if docii in docs:
                        # remove duplicates
                        continue

                    cii=meta_dict[docii]['citationkey']
                    if cii=='':
                        cii=str(docii)
                    citationkeysii.append(cii)
                    jobsii.append((n, meta_dict[docii], prefix))
                    n+=1
                    docs.append(docii)

                #---------------Sort by citationkey---------------
                jobsii=[x for _,x in sorted(zip(citationkeysii,jobsii))]
                citationkeysii=sorted(citationkeysii)
                job_list.extend(jobsii)
                citationkeys.extend(citationkeysii)

        #------------------Run in thread------------------
        thread_run_dialog=ThreadRunDialog(risparse.metaDictToRIS, job_list,
                show_message='Processing...', max_threads=1, get_results=False,
                close_on_finish=False,
                progressbar_style='classic',
                post_process_func=self.saveRIS,
                post_process_func_args=(folders, manner, fname, folder_data,
                    meta_dict,citationkeys),
                parent=self)

        thread_run_dialog.exec_()

        return


    def saveRIS(self, results, folders, manner, fname, folder_data, meta_dict,
            citationkeys):
        """Save exported RIS text to files

        Args:
            results (list): return values of metaDictToRiIS().
            folders (list): list of folders in the format [(name, folderid), ].
            manner (str): saving manner, 'All in one': save all RIS entries
                          in a single file.
                          'Per folder': group by folder.
                          'Per document': per doc.
            fname (str): if manner=='All in one', the abspath to the output
                         ris file. Otherwise, the abspath to folder to save
                         output ris files.
            folder_data (dict): documents in each folder. keys: folder id in str,
                                values: list of doc ids. To determine when
                                a folder gets all its docs exported.
            meta_dict (dict): meta data of all documents. keys: docid,
                              values: DocMeta dict.
            citationkeys (list): list of the citationkeys for each doc. Used
                                 as the file name in 'Per document' mode.

        """
        faillist=[]

        if manner=='All in one':
            text=''
            for recii,jobii,textii,docii in results:
                if recii==0:
                    text=text+textii
                elif recii==1:
                    faillist.append(docii)

            with open(fname,'w') as fout:
                fout.write(text)

            LOGGER.info('Saved combined outputs to file %s' %fname)

        elif manner=='Per document':
            for recii,jobii,textii,docii in results:
                if recii==0:
                    citationkey=citationkeys[jobii]
                    fnameii='%s.ris' %citationkey
                    fnameii=os.path.join(fname,fnameii)
                    fnameii=autoRename(fnameii)

                    with open(fnameii,'w') as fout:
                        fout.write(textii)

                    LOGGER.info('Saved per doc output to file %s' %fnameii)

                elif recii==1:
                    faillist.append(docii)

        elif manner=='Per folder':

            #-------------Prepare result counting-------------
            folder_counts={}
            folder_results={}
            for foldernameii,fidii in folders:
                folder_results[fidii]=[]
                folder_counts[fidii]=0

            for recii,jobii,textii,docii in results:
                if recii==1:
                    faillist.append(docii)

                for fnamejj,fidjj in folders:
                    if docii in folder_data[fidjj]:
                        folder_counts[fidjj]+=1
                        if recii==0:
                            folder_results[fidjj].append(textii)

                        # save if folder got all its doc processed, even if failed
                        if folder_counts[fidjj]==len(folder_data[fidjj]):

                            fnamejj=os.path.join(fname,'%s.ris' %fnamejj)
                            text='\n'.join(folder_results[fidjj])

                            with open(fnamejj,'w') as fout:
                                fout.write(text)

                            LOGGER.info('Folder %s got all data. Saved.'\
                                    %fnamejj)

        #-----------------Show failed jobs-----------------
        if len(faillist)>0:
            fail_entries=[]
            for docii in faillist:
                metaii=meta_dict[docii]
                entryii='* %s_%s_%s' %(', '.join(metaii['authors_l']),
                        metaii['year'],
                        metaii['title'])
                fail_entries.append(entryii)

                LOGGER.warning('Failed export job %s' %entryii)

            msg=FailDialog()
            msg.setText('Oopsie.')
            msg.setInformativeText('Failed to export some entries.')
            msg.setDetailedText('\n'.join(fail_entries))
            msg.create_fail_summary.connect(lambda:\
                    self.parent.main_frame.createFailFolder('RIS export',
                        faillist))
            msg.exec_()

        return



    def doExport(self):

        LOGGER.info('task = %s' %self.current_task)

        if self.current_task=='copy_file':
            self.doExportFiles()
        elif self.current_task=='bib_export':
            self.doBibExport()
        elif self.current_task=='ris_export':
            self.doRISExport()
        elif self.current_task=='zotero_export':
            pass

        return


    def popUpChooseFolder(self):

        msg=QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        msg.setWindowTitle('Input Needed')
        msg.setText("Choose at least one folder to process.")
        msg.exec_()

        return




import os
import shutil
import logging
from collections import OrderedDict
from PyQt5 import QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal, pyqtSlot, QRegExp
from PyQt5.QtWidgets import QDialogButtonBox
from PyQt5.QtGui import QFont, QRegExpValidator
import resources
from .. import sqlitedb
from .. import bibparse
from .. import risparse
from ..tools import getHLine, createFolderTree, iterTreeWidgetItems, autoRename
from .threadrun_dialog import ThreadRunDialog
from .fail_dialog import FailDialog

LOGGER=logging.getLogger(__name__)




class CreateNewLibDialog(QtWidgets.QDialog):
    def __init__(self,settings,parent):

        super(CreateNewLibDialog,self).__init__(parent=parent)

        self.settings=settings
        self.parent=parent

        self.resize(500,250)
        self.label_color='color: rgb(0,0,140); background-color: rgb(235,235,240)'
        self.title_label_font=QFont('Serif',10,QFont.Bold)
        self.setWindowTitle('Create New Library')
        self.setWindowModality(Qt.ApplicationModal)

        v_layout=QtWidgets.QVBoxLayout(self)

        #---------------------Lib name---------------------
        label=QtWidgets.QLabel('Name your library')
        #label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)
        v_layout.addWidget(label)

        self.lib_name_le=QtWidgets.QLineEdit()
        regex=QRegExp("[a-z-A-Z_\d]+")
        validator = QRegExpValidator(regex)
        self.lib_name_le.setValidator(validator)

        label=QtWidgets.QLabel('(Only alphanumeric characters and "-", "_" are allowed)')
        v_layout.addWidget(label)

        #ha=QtWidgets.QHBoxLayout()
        v_layout.addWidget(self.lib_name_le)
        #ha.addWidget(label)

        #v_layout.addLayout(ha)
        v_layout.addWidget(getHLine())

        #-------------------Save folder-------------------
        storage_folder=self.settings.value('saving/storage_folder',str)

        label=QtWidgets.QLabel('Storage Folder')
        #label.setStyleSheet(self.label_color)
        label.setFont(self.title_label_font)
        v_layout.addWidget(label)

        self.folder_le=QtWidgets.QLineEdit(self)
        #self.folder_le.setPlaceholderText(storage_folder)
        self.folder_le.setText(storage_folder)

        self.open_button=QtWidgets.QPushButton(self)
        self.open_button.setText('Open')
        self.open_button.clicked.connect(lambda: self.dirChooseButtonClicked(
            self.folder_le))

        ha=QtWidgets.QHBoxLayout()
        ha.addWidget(self.folder_le)
        ha.addWidget(self.open_button)
        v_layout.addLayout(ha)

        #------------------Dialog buttons------------------
        self.buttons=QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel,
                Qt.Horizontal, self)

        self.buttons.accepted.connect(self.okClicked)
        self.buttons.rejected.connect(self.reject)

        v_layout.addStretch()
        v_layout.addWidget(self.buttons)


    @pyqtSlot(QtWidgets.QLineEdit)
    def dirChooseButtonClicked(self, le):

        storage_folder=self.settings.value('saving/storage_folder',str)
        fname = QtWidgets.QFileDialog.getExistingDirectory(self,
                'Choose a folder to save library files',
                storage_folder)

        if fname:
            print('# <dirChooseButtonClicked>: fname = ',fname)
            #LOGGER.info('Choose file name %s' %fname)
            le.setText(fname)

        return


    @pyqtSlot()
    def okClicked(self):

        lib_name=self.lib_name_le.text()
        fname=self.folder_le.text()
        if lib_name=='':
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setWindowTitle('Input needed')
            msg.setText("Library name can't be empty")
            msg.exec_()
            return

        if fname=='':
            msg=QtWidgets.QMessageBox()
            msg.setIcon(QtWidgets.QMessageBox.Information)
            msg.setWindowTitle('Input needed')
            msg.setText("Choose a folder to save library files.")
            msg.exec_()
            return

        self.accept()


